import json

import pytest

from great_expectations.core import ExpectationConfiguration
from great_expectations.core.expectation_validation_result import (
    ExpectationValidationResult,
)
from great_expectations.expectations.registry import get_renderer_impl
from great_expectations.render import (
    LegacyDiagnosticRendererType,
    RenderedComponentContent,
    RenderedStringTemplateContent,
)
from great_expectations.render.renderer.content_block import (
    ValidationResultsTableContentBlockRenderer,
)


@pytest.fixture
def evr_failed_with_exception():
    return ExpectationValidationResult(
        success=False,
        exception_info={
            "raised_exception": True,
            "exception_message": "Invalid partition object.",
            "exception_traceback": 'Traceback (most recent call last):\n  File "/great_expectations/great_expectations/data_asset/data_asset.py", line 216, in wrapper\n    return_obj = func(self, **evaluation_args)\n  File "/great_expectations/great_expectations/dataset/dataset.py", line 106, in inner_wrapper\n    evaluation_result = func(self, column, *args, **kwargs)\n  File "/great_expectations/great_expectations/dataset/dataset.py", line 3381, in expect_column_kl_divergence_to_be_less_than\n    raise ValueError("Invalid partition object.")\nValueError: Invalid partition object.\n',
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_kl_divergence_to_be_less_than",
            kwargs={
                "column": "live",
                "partition_object": None,
                "threshold": None,
                "result_format": "SUMMARY",
            },
            meta={"BasicDatasetProfiler": {"confidence": "very low"}},
        ),
    )


@pytest.fixture()
def evr_id_pk_basic_sql() -> ExpectationValidationResult:
    return ExpectationValidationResult(
        success=False,
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={
                "batch_id": "cb8e223838fcdb055f6cccad2af592ae",
                "column": "animals",
                "value_set": ["cat", "fish", "dog"],
            },
            meta={},
        ),
        exception_info={
            "exception_message": None,
            "exception_traceback": None,
            "raised_exception": False,
        },
        result={
            "element_count": 6,
            "missing_count": 0,
            "missing_percent": 0.0,
            "partial_unexpected_counts": [
                {"count": 1, "value": "giraffe"},
                {"count": 1, "value": "lion"},
                {"count": 1, "value": "zebra"},
            ],
            "partial_unexpected_index_list": [
                {"animals": "giraffe", "pk_1": 3, "pk_2": "three"},
                {"animals": "lion", "pk_1": 4, "pk_2": "four"},
                {"animals": "zebra", "pk_1": 5, "pk_2": "five"},
            ],
            "partial_unexpected_list": ["giraffe", "lion", "zebra"],
            "unexpected_count": 3,
            "unexpected_index_column_names": ["pk_1", "pk_2"],
            "unexpected_index_list": [
                {"animals": "giraffe", "pk_1": 3, "pk_2": "three"},
                {"animals": "lion", "pk_1": 4, "pk_2": "four"},
                {"animals": "zebra", "pk_1": 5, "pk_2": "five"},
            ],
            "unexpected_index_query": "SELECT animals, pk_1, pk_2 \nFROM animal_names \nWHERE animals IS NOT NULL AND (animals NOT IN ('cat', 'fish', 'dog'))",
            "unexpected_list": ["giraffe", "lion", "zebra"],
            "unexpected_percent": 50.0,
            "unexpected_percent_nonmissing": 50.0,
            "unexpected_percent_total": 50.0,
        },
    )


@pytest.fixture()
def evr_id_pk_basic_pandas() -> ExpectationValidationResult:
    return ExpectationValidationResult(
        success=False,
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={
                "batch_id": "cb8e223838fcdb055f6cccad2af592ae",
                "column": "animals",
                "value_set": ["cat", "fish", "dog"],
            },
            meta={},
        ),
        exception_info={
            "exception_message": None,
            "exception_traceback": None,
            "raised_exception": False,
        },
        result={
            "element_count": 6,
            "missing_count": 0,
            "missing_percent": 0.0,
            "partial_unexpected_counts": [
                {"count": 1, "value": "giraffe"},
                {"count": 1, "value": "lion"},
                {"count": 1, "value": "zebra"},
            ],
            "partial_unexpected_index_list": [3, 4, 5],
            "partial_unexpected_list": ["giraffe", "lion", "zebra"],
            "unexpected_count": 3,
            "unexpected_index_list": [3, 4, 5],
            "unexpected_index_query": [3, 4, 5],
            "unexpected_list": ["giraffe", "lion", "zebra"],
            "unexpected_percent": 50.0,
            "unexpected_percent_nonmissing": 50.0,
            "unexpected_percent_total": 50.0,
        },
    )


def test_ValidationResultsTableContentBlockRenderer_generate_expectation_row_with_errored_expectation(
    evr_failed_with_exception,
):
    result = ValidationResultsTableContentBlockRenderer.render(
        [evr_failed_with_exception]
    ).to_json_dict()
    expected_result = {
        "content_block_type": "table",
        "styling": {
            "body": {"classes": ["table"]},
            "classes": ["ml-2", "mr-2", "mt-0", "mb-0", "table-responsive"],
        },
        "table": [
            [
                {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "❗"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": [
                                        "fas",
                                        "fa-exclamation-triangle",
                                        "text-warning",
                                    ],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                },
                [
                    {
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": "$column can match any distribution.",
                            "params": {
                                "column": "live",
                                "partition_object": None,
                                "threshold": None,
                                "result_format": "SUMMARY",
                                "row_condition": None,
                                "condition_parser": None,
                            },
                        },
                    },
                    {
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": "\n\n$expectation_type raised an exception:\n$exception_message",
                            "params": {
                                "expectation_type": "expect_column_kl_divergence_to_be_less_than",
                                "exception_message": "Invalid partition object.",
                            },
                            "tag": "strong",
                            "styling": {
                                "classes": ["text-danger"],
                                "params": {
                                    "exception_message": {"tag": "code"},
                                    "expectation_type": {
                                        "classes": ["badge", "badge-danger", "mb-2"]
                                    },
                                },
                            },
                        },
                    },
                    {
                        "content_block_type": "collapse",
                        "collapse_toggle_link": "Show exception traceback...",
                        "collapse": [
                            {
                                "content_block_type": "string_template",
                                "string_template": {
                                    "template": 'Traceback (most recent call last):\n  File "/great_expectations/great_expectations/data_asset/data_asset.py", line 216, in wrapper\n    return_obj = func(self, **evaluation_args)\n  File "/great_expectations/great_expectations/dataset/dataset.py", line 106, in inner_wrapper\n    evaluation_result = func(self, column, *args, **kwargs)\n  File "/great_expectations/great_expectations/dataset/dataset.py", line 3381, in expect_column_kl_divergence_to_be_less_than\n    raise ValueError("Invalid partition object.")\nValueError: Invalid partition object.\n',
                                    "tag": "code",
                                },
                            }
                        ],
                        "inline_link": False,
                    },
                ],
                "--",
            ]
        ],
        "header_row": ["Status", "Expectation", "Observed Value"],
        "header_row_options": {"Status": {"sortable": True}},
        "table_options": {"search": True, "icon-size": "sm"},
    }
    assert result == expected_result


def test_ValidationResultsTableContentBlockRenderer_render(
    titanic_profiled_name_column_evrs,
):
    validation_results_table = ValidationResultsTableContentBlockRenderer.render(
        titanic_profiled_name_column_evrs
    )

    assert isinstance(validation_results_table, RenderedComponentContent)
    assert validation_results_table.content_block_type == "table"
    assert len(validation_results_table.table) == 6
    assert validation_results_table.header_row == [
        "Status",
        "Expectation",
        "Observed Value",
    ]
    assert validation_results_table.styling == {
        "body": {"classes": ["table"]},
        "classes": ["ml-2", "mr-2", "mt-0", "mb-0", "table-responsive"],
    }
    assert json.dumps(validation_results_table.to_json_dict()).count("$icon") == 6


def test_ValidationResultsTableContentBlockRenderer_get_custom_columns(evr_success):
    assert (
        ValidationResultsTableContentBlockRenderer._get_custom_columns([evr_success])
        == []
    )

    evr_success.expectation_config.kwargs["meta_properties_to_render"] = {}
    assert (
        ValidationResultsTableContentBlockRenderer._get_custom_columns([evr_success])
        == []
    )

    evr_success.expectation_config.kwargs["meta_properties_to_render"] = {
        "doesntmatterone": "doesntmatter",
        "doesntmattertwo": "doesntmatter",
    }
    assert ValidationResultsTableContentBlockRenderer._get_custom_columns(
        [evr_success]
    ) == ["doesntmatterone", "doesntmattertwo"]


def test_ValidationResultsTableContentBlockRenderer_get_content_block_fn(evr_success):
    content_block_fn = ValidationResultsTableContentBlockRenderer._get_content_block_fn(
        "expect_table_row_count_to_be_between"
    )
    content_block_fn_output = content_block_fn(result=evr_success)

    content_block_fn_expected_output = [
        [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "✅"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": [
                                        "fas",
                                        "fa-check-circle",
                                        "text-success",
                                    ],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                    "styling": {
                        "parent": {
                            "classes": ["hide-succeeded-validation-target-child"]
                        }
                    },
                }
            ),
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Must have greater than or equal to $min_value rows.",
                        "params": {
                            "min_value": 0,
                            "max_value": None,
                            "result_format": "SUMMARY",
                            "strict_max": None,
                            "strict_min": None,
                        },
                        "styling": None,
                    },
                }
            ),
            "1,313",
        ]
    ]
    assert content_block_fn_output == content_block_fn_expected_output

    content_block_fn_expected_output = [
        [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "✅"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": [
                                        "fas",
                                        "fa-check-circle",
                                        "text-success",
                                    ],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                    "styling": {
                        "parent": {
                            "classes": ["hide-succeeded-validation-target-child"]
                        }
                    },
                }
            ),
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Must have greater than or equal to $min_value rows.",
                        "params": {
                            "meta_properties_to_render": {},
                            "min_value": 0,
                            "max_value": None,
                            "result_format": "SUMMARY",
                            "strict_max": None,
                            "strict_min": None,
                        },
                        "styling": None,
                    },
                }
            ),
            "1,313",
        ]
    ]

    evr_success.expectation_config.kwargs["meta_properties_to_render"] = {}
    content_block_fn_output = content_block_fn(result=evr_success)
    assert (
        content_block_fn_output[0][0].to_json_dict()
        == content_block_fn_expected_output[0][0].to_json_dict()
    )
    assert (
        content_block_fn_output[0][1].to_json_dict()
        == content_block_fn_expected_output[0][1].to_json_dict()
    )

    assert content_block_fn_output == content_block_fn_expected_output

    evr_success.expectation_config.kwargs["meta_properties_to_render"] = {
        "property_that_doesnt_exist": "property"
    }
    content_block_fn_expected_output = [
        [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "✅"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": [
                                        "fas",
                                        "fa-check-circle",
                                        "text-success",
                                    ],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                    "styling": {
                        "parent": {
                            "classes": ["hide-succeeded-validation-target-child"]
                        }
                    },
                }
            ),
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Must have greater than or equal to $min_value rows.",
                        "params": {
                            "meta_properties_to_render": {
                                "property_that_doesnt_exist": "property"
                            },
                            "min_value": 0,
                            "max_value": None,
                            "result_format": "SUMMARY",
                            "strict_max": None,
                            "strict_min": None,
                        },
                        "styling": None,
                    },
                }
            ),
            "1,313",
            ["N/A"],
        ]
    ]
    content_block_fn_output = content_block_fn(result=evr_success)
    assert content_block_fn_output == content_block_fn_expected_output

    evr_success.expectation_config.meta = {
        "attributes": {"property": 5, "nested": {"property": "this is nested"}}
    }
    evr_success.expectation_config.kwargs["meta_properties_to_render"] = {
        "property_that_exists": "property",
        "other existing prop": "nested.property",
    }

    content_block_fn_expected_output = [
        [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "✅"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": [
                                        "fas",
                                        "fa-check-circle",
                                        "text-success",
                                    ],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                    "styling": {
                        "parent": {
                            "classes": ["hide-succeeded-validation-target-child"]
                        }
                    },
                }
            ),
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Must have greater than or equal to $min_value rows.",
                        "params": {
                            "meta_properties_to_render": {
                                "property_that_exists": "property",
                                "other existing prop": "nested.property",
                            },
                            "min_value": 0,
                            "max_value": None,
                            "result_format": "SUMMARY",
                            "strict_max": None,
                            "strict_min": None,
                        },
                        "styling": None,
                    },
                }
            ),
            "1,313",
            ["this is nested"],
            [5],
        ]
    ]
    content_block_fn_output = content_block_fn(result=evr_success)
    assert content_block_fn_output == content_block_fn_expected_output


@pytest.mark.filterwarnings("ignore:V2 API style custom rendering*:DeprecationWarning")
def test_ValidationResultsTableContentBlockRenderer_get_content_block_fn_with_v2_api_style_custom_rendering():
    """Test backwards support for custom expectation rendering with the V2 API as described at
    https://docs.greatexpectations.io/en/latest/reference/spare_parts/data_docs_reference.html#customizing-data-docs.
    """

    custom_expectation_template = "custom_expectation_template"
    custom_expectation_observed_value = "custom_expectation_observed_value"

    class ValidationResultsTableContentBlockRendererWithV2ApiStyleCustomExpectations(
        ValidationResultsTableContentBlockRenderer
    ):
        @classmethod
        def expect_custom_expectation_written_in_v2_api_style(
            cls, expectation, styling=None, include_column_name: bool = True
        ):
            return [
                RenderedStringTemplateContent(
                    content_block_type="string_template",
                    string_template={
                        "template": custom_expectation_template,
                        "params": expectation.kwargs,
                        "styling": styling,
                    },
                )
            ]

    evr = ExpectationValidationResult(
        success=True,
        result={
            "observed_value": custom_expectation_observed_value,
        },
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_custom_expectation",
            kwargs={"column": "a_column_name", "result_format": "SUMMARY"},
        ),
    )

    content_block_fn = ValidationResultsTableContentBlockRendererWithV2ApiStyleCustomExpectations._get_content_block_fn(
        "expect_custom_expectation_written_in_v2_api_style"
    )
    content_block_fn_output = content_block_fn(result=evr)

    content_block_fn_expected_output = [
        [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "✅"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": [
                                        "fas",
                                        "fa-check-circle",
                                        "text-success",
                                    ],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                    "styling": {
                        "parent": {
                            "classes": ["hide-succeeded-validation-target-child"]
                        }
                    },
                }
            ),
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": custom_expectation_template,
                        "params": {
                            "column": "a_column_name",
                            "result_format": "SUMMARY",
                        },
                        "styling": None,
                    },
                }
            ),
            custom_expectation_observed_value,
        ]
    ]
    assert content_block_fn_output == content_block_fn_expected_output


def test_ValidationResultsTableContentBlockRenderer_get_observed_value(evr_success):
    evr_no_result_key = ExpectationValidationResult(
        success=True,
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_table_row_count_to_be_between",
            kwargs={"min_value": 0, "max_value": None, "result_format": "SUMMARY"},
        ),
    )

    evr_expect_column_values_to_not_be_null = ExpectationValidationResult(
        success=True,
        result={
            "element_count": 1313,
            "unexpected_count": 1050,
            "unexpected_percent": 79.96953541508,
            "partial_unexpected_list": [],
        },
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_not_be_null",
            kwargs={"column": "Unnamed: 0", "mostly": 0.5, "result_format": "SUMMARY"},
        ),
    )

    evr_expect_column_values_to_be_null = ExpectationValidationResult(
        success=True,
        result={
            "element_count": 1313,
            "unexpected_count": 0,
            "unexpected_percent": 0.0,
            "partial_unexpected_list": [],
        },
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_null",
            kwargs={"column": "Unnamed: 0", "mostly": 0.5, "result_format": "SUMMARY"},
        ),
    )

    evr_success_zero = ExpectationValidationResult(
        success=True,
        result={"observed_value": 0},
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_table_row_count_to_be_between",
            kwargs={"min_value": 0, "max_value": None, "result_format": "SUMMARY"},
        ),
    )

    # test _get_observed_value when evr.result["observed_value"] exists
    output_1 = get_renderer_impl(
        object_name=evr_success.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE,
    )[1](result=evr_success)
    assert output_1 == "1,313"
    # test _get_observed_value when evr.result does not exist
    output_2 = get_renderer_impl(
        object_name=evr_no_result_key.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE,
    )[1](result=evr_no_result_key)
    assert output_2 == "--"
    # test _get_observed_value for expect_column_values_to_not_be_null expectation type
    output_3 = get_renderer_impl(
        object_name=evr_expect_column_values_to_not_be_null.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE,
    )[1](result=evr_expect_column_values_to_not_be_null)
    assert output_3 == "≈20.03% not null"
    # test _get_observed_value for expect_column_values_to_be_null expectation type
    output_4 = get_renderer_impl(
        object_name=evr_expect_column_values_to_be_null.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE,
    )[1](result=evr_expect_column_values_to_be_null)
    assert output_4 == "100% null"
    # test _get_observed_value to be 0
    output_5 = get_renderer_impl(
        object_name=evr_success_zero.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE,
    )[1](result=evr_success_zero)
    assert output_5 == "0"


def test_ValidationResultsTableContentBlockRenderer_get_unexpected_statement(
    evr_success, evr_failed
):
    evr_no_result = ExpectationValidationResult(
        success=True,
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_table_row_count_to_be_between",
            kwargs={"min_value": 0, "max_value": None, "result_format": "SUMMARY"},
        ),
    )
    evr_failed_no_unexpected_count = ExpectationValidationResult(
        success=False,
        result={
            "element_count": 1313,
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_percent": 0.2284843869002285,
            "unexpected_percent_nonmissing": 0.2284843869002285,
            "partial_unexpected_list": [
                "Daly, Mr Peter Denis ",
                "Barber, Ms ",
                "Geiger, Miss Emily ",
            ],
            "partial_unexpected_index_list": [77, 289, 303],
            "partial_unexpected_counts": [
                {"value": "Barber, Ms ", "count": 1},
                {"value": "Daly, Mr Peter Denis ", "count": 1},
                {"value": "Geiger, Miss Emily ", "count": 1},
            ],
        },
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_not_match_regex",
            kwargs={
                "column": "Name",
                "regex": "^\\s+|\\s+$",
                "result_format": "SUMMARY",
            },
        ),
    )

    # test for succeeded evr
    output_1 = get_renderer_impl(
        object_name=evr_success.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_STATEMENT,
    )[1](result=evr_success)
    assert output_1 == []

    # test for failed evr
    output_2 = get_renderer_impl(
        object_name=evr_failed.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_STATEMENT,
    )[1](result=evr_failed)
    assert output_2 == [
        RenderedStringTemplateContent(
            **{
                "content_block_type": "string_template",
                "string_template": {
                    "template": "\n\n$unexpected_count unexpected values found. $unexpected_percent of $element_count total rows.",
                    "params": {
                        "unexpected_count": "3",
                        "unexpected_percent": "≈0.2285%",
                        "element_count": "1,313",
                    },
                    "tag": "strong",
                    "styling": {"classes": ["text-danger"]},
                },
            }
        )
    ]

    # test for evr with no "result" key
    output_3 = get_renderer_impl(
        object_name=evr_no_result.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_STATEMENT,
    )[1](result=evr_no_result)
    print(json.dumps(output_3, indent=2))
    assert output_3 == []

    # test for evr with no unexpected count
    output_4 = get_renderer_impl(
        object_name=evr_failed_no_unexpected_count.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_STATEMENT,
    )[1](result=evr_failed_no_unexpected_count)
    print(output_4)
    assert output_4 == []

    # test for evr with exception
    evr_failed_exception = ExpectationValidationResult(
        success=False,
        exception_info={
            "raised_exception": True,
            "exception_message": "Unrecognized column: not_a_real_column",
            "exception_traceback": "Traceback (most recent call last):\n...more_traceback...",
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_not_match_regex",
            kwargs={
                "column": "Name",
                "regex": "^\\s+|\\s+$",
                "result_format": "SUMMARY",
            },
        ),
    )

    output_5 = get_renderer_impl(
        object_name=evr_failed_exception.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_STATEMENT,
    )[1](result=evr_failed_exception)
    output_5 = [content.to_json_dict() for content in output_5]
    expected_output_5 = [
        {
            "content_block_type": "string_template",
            "string_template": {
                "template": "\n\n$expectation_type raised an exception:\n$exception_message",
                "params": {
                    "expectation_type": "expect_column_values_to_not_match_regex",
                    "exception_message": "Unrecognized column: not_a_real_column",
                },
                "tag": "strong",
                "styling": {
                    "classes": ["text-danger"],
                    "params": {
                        "exception_message": {"tag": "code"},
                        "expectation_type": {
                            "classes": ["badge", "badge-danger", "mb-2"]
                        },
                    },
                },
            },
        },
        {
            "content_block_type": "collapse",
            "collapse_toggle_link": "Show exception traceback...",
            "collapse": [
                {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Traceback (most recent call last):\n...more_traceback...",
                        "tag": "code",
                    },
                }
            ],
            "inline_link": False,
        },
    ]
    assert output_5 == expected_output_5


def test_ValidationResultsTableContentBlockRenderer_get_unexpected_table(evr_success):
    evr_failed_no_result = ExpectationValidationResult(
        success=False,
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={
                "column": "Unnamed: 0",
                "value_set": [],
                "result_format": "SUMMARY",
            },
        ),
    )

    evr_failed_no_unexpected_list_or_counts = ExpectationValidationResult(
        success=False,
        result={
            "element_count": 1313,
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_count": 1313,
            "unexpected_percent": 100.0,
            "unexpected_percent_nonmissing": 100.0,
        },
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={
                "column": "Unnamed: 0",
                "value_set": [],
                "result_format": "SUMMARY",
            },
        ),
    )

    evr_failed_partial_unexpected_list = ExpectationValidationResult(
        success=False,
        result={
            "element_count": 1313,
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_count": 1313,
            "unexpected_percent": 100.0,
            "unexpected_percent_nonmissing": 100.0,
            "partial_unexpected_list": [
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                19,
                20,
            ],
        },
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={
                "column": "Unnamed: 0",
                "value_set": [],
                "result_format": "SUMMARY",
            },
        ),
    )

    evr_failed_partial_unexpected_counts = ExpectationValidationResult(
        success=False,
        result={
            "element_count": 1313,
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_count": 1313,
            "unexpected_percent": 100.0,
            "unexpected_percent_nonmissing": 100.0,
            "partial_unexpected_list": [
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                19,
                20,
            ],
            "partial_unexpected_index_list": [
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                19,
            ],
            "partial_unexpected_counts": [
                {"value": 1, "count": 1},
                {"value": 2, "count": 1},
                {"value": 3, "count": 1},
                {"value": 4, "count": 1},
                {"value": 5, "count": 1},
                {"value": 6, "count": 1},
                {"value": 7, "count": 1},
                {"value": 8, "count": 1},
                {"value": 9, "count": 1},
                {"value": 10, "count": 1},
                {"value": 11, "count": 1},
                {"value": 12, "count": 1},
                {"value": 13, "count": 1},
                {"value": 14, "count": 1},
                {"value": 15, "count": 1},
                {"value": 16, "count": 1},
                {"value": 17, "count": 1},
                {"value": 18, "count": 1},
                {"value": 19, "count": 1},
                {"value": 20, "count": 1},
            ],
        },
        exception_info={
            "raised_exception": False,
            "exception_message": None,
            "exception_traceback": None,
        },
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={
                "column": "Unnamed: 0",
                "value_set": [],
                "result_format": "SUMMARY",
            },
        ),
    )

    # test for succeeded evr
    output_1 = get_renderer_impl(
        object_name=evr_success.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_success)
    assert output_1 is None

    # test for failed evr with no "result" key
    output_2 = get_renderer_impl(
        object_name=evr_failed_no_result.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_failed_no_result)
    assert output_2 is None

    # test for failed evr with no unexpected list or unexpected counts
    output_3 = get_renderer_impl(
        object_name=evr_failed_no_unexpected_list_or_counts.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_failed_no_unexpected_list_or_counts)
    assert output_3 is None

    # test for failed evr with partial unexpected list
    output_4 = get_renderer_impl(
        object_name=evr_failed_partial_unexpected_list.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_failed_partial_unexpected_list)
    assert output_4[0].to_json_dict() == {
        "content_block_type": "table",
        "table": [
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
            [8],
            [9],
            [10],
            [11],
            [12],
            [13],
            [14],
            [15],
            [16],
            [17],
            [18],
            [19],
            [20],
        ],
        "header_row": ["Sampled Unexpected Values"],
        "styling": {"body": {"classes": ["table-bordered", "table-sm", "mt-3"]}},
    }

    # test for failed evr with partial unexpected counts
    output_5 = get_renderer_impl(
        object_name=evr_failed_partial_unexpected_counts.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_failed_partial_unexpected_counts)
    assert output_5[0].to_json_dict() == {
        "content_block_type": "table",
        "header_row": ["Sampled Unexpected Values", "Count"],
        "styling": {"body": {"classes": ["table-bordered", "table-sm", "mt-3"]}},
        "table": [
            [1, 1],
            [2, 1],
            [3, 1],
            [4, 1],
            [5, 1],
            [6, 1],
            [7, 1],
            [8, 1],
            [9, 1],
            [10, 1],
            [11, 1],
            [12, 1],
            [13, 1],
            [14, 1],
            [15, 1],
            [16, 1],
            [17, 1],
            [18, 1],
            [19, 1],
            [20, 1],
        ],
    }


def test_ValidationResultsTableContentBlockRenderer_get_status_cell(
    evr_failed_with_exception, evr_success, evr_failed
):
    # test for failed evr with exception
    output_1 = get_renderer_impl(
        object_name=evr_failed_with_exception.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.STATUS_ICON,
    )[1](result=evr_failed_with_exception)
    assert output_1.to_json_dict() == {
        "content_block_type": "string_template",
        "string_template": {
            "template": "$icon",
            "params": {"icon": "", "markdown_status_icon": "❗"},
            "styling": {
                "params": {
                    "icon": {
                        "classes": ["fas", "fa-exclamation-triangle", "text-warning"],
                        "tag": "i",
                    }
                }
            },
        },
    }

    # test for succeeded evr
    output_2 = get_renderer_impl(
        object_name=evr_success.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.STATUS_ICON,
    )[1](result=evr_success)
    assert output_2.to_json_dict() == {
        "content_block_type": "string_template",
        "string_template": {
            "template": "$icon",
            "params": {"icon": "", "markdown_status_icon": "✅"},
            "styling": {
                "params": {
                    "icon": {
                        "classes": ["fas", "fa-check-circle", "text-success"],
                        "tag": "i",
                    }
                }
            },
        },
        "styling": {"parent": {"classes": ["hide-succeeded-validation-target-child"]}},
    }

    # test for failed evr
    output_3 = get_renderer_impl(
        object_name=evr_failed.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.STATUS_ICON,
    )[1](result=evr_failed)
    assert output_3.to_json_dict() == {
        "content_block_type": "string_template",
        "string_template": {
            "template": "$icon",
            "params": {"icon": "", "markdown_status_icon": "❌"},
            "styling": {
                "params": {
                    "icon": {"tag": "i", "classes": ["fas", "fa-times", "text-danger"]}
                }
            },
        },
    }


def test_ValidationResultsTableContentBlockRenderer_get_unexpected_table_no_id_pk_pandas():
    evr_no_id_pk_pandas = ExpectationValidationResult(
        success=False,
        expectation_config=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={
                "batch_id": "cb8e223838fcdb055f6cccad2af592ae",
                "column": "animals",
                "value_set": ["cat", "fish", "dog"],
            },
            meta={},
        ),
        result={
            "element_count": 6,
            "missing_count": 0,
            "missing_percent": 0.0,
            "partial_unexpected_counts": [
                {"count": 1, "value": "giraffe"},
                {"count": 1, "value": "lion"},
                {"count": 1, "value": "zebra"},
            ],
            "partial_unexpected_index_list": [3, 4, 5],
            "partial_unexpected_list": ["giraffe", "lion", "zebra"],
            "unexpected_count": 3,
            "unexpected_index_list": [3, 4, 5],
            "unexpected_list": ["giraffe", "lion", "zebra"],
            "unexpected_percent": 50.0,
            "unexpected_percent_nonmissing": 50.0,
            "unexpected_percent_total": 50.0,
        },
        exception_info={
            "exception_message": None,
            "exception_traceback": None,
            "raised_exception": False,
        },
    )
    rendered_value = get_renderer_impl(
        object_name=evr_no_id_pk_pandas.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_no_id_pk_pandas)
    assert rendered_value[0].to_json_dict() == {
        "content_block_type": "table",
        "styling": {"body": {"classes": ["table-bordered", "table-sm", "mt-3"]}},
        "table": [["giraffe", 1, "3"], ["lion", 1, "4"], ["zebra", 1, "5"]],
        "header_row": ["Unexpected Value", "Count", "Index"],
    }


def test_ValidationResultsTableContentBlockRenderer_get_unexpected_table_with_id_pk_pandas_and_query(
    evr_id_pk_basic_pandas,
):
    rendered_value = get_renderer_impl(
        object_name=evr_id_pk_basic_pandas.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_id_pk_basic_pandas)
    assert rendered_value[0].to_json_dict() == {
        "content_block_type": "table",
        "header_row": ["Unexpected Value", "Count", "Index"],
        "styling": {"body": {"classes": ["table-bordered", "table-sm", "mt-3"]}},
        "table": [["giraffe", 1, "3"], ["lion", 1, "4"], ["zebra", 1, "5"]],
    }
    assert rendered_value[1].to_json_dict() == {
        "collapse": [
            {
                "content_block_type": "string_template",
                "string_template": {"tag": "code", "template": "[3, 4, 5]"},
            }
        ],
        "collapse_toggle_link": "To retrieve all unexpected values...",
        "content_block_type": "collapse",
        "inline_link": False,
    }


def test_ValidationResultsTableContentBlockRenderer_get_unexpected_table_with_id_pk_pandas_with_sampled_table(
    evr_id_pk_basic_pandas,
):
    evr_id_pk_pandas = evr_id_pk_basic_pandas
    new_index = [3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
    new_unexpected_list = [
        "giraffe",
        "lion",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
    ]
    evr_id_pk_pandas.result["partial_unexpected_index_list"] = new_index
    evr_id_pk_pandas.result["partial_unexpected_list"] = new_unexpected_list
    evr_id_pk_pandas.result["unexpected_count"] = 100
    evr_id_pk_pandas.result["unexpected_index_list"] = new_index
    evr_id_pk_pandas.result["unexpected_list"] = new_unexpected_list

    rendered_value = get_renderer_impl(
        object_name=evr_id_pk_pandas.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_id_pk_pandas)
    assert rendered_value[0].to_json_dict() == {
        "content_block_type": "table",
        "header_row": ["Sampled Unexpected Values", "Count", "Index"],
        "styling": {"body": {"classes": ["table-bordered", "table-sm", "mt-3"]}},
        "table": [
            ["giraffe", 1, "3"],
            ["lion", 1, "4"],
            ["zebra", 11, "5, 6, 7, 8, 9, 10, 11, 12, 13, 14, ..."],
        ],
    }


def test_ValidationResultsTableContentBlockRenderer_get_unexpected_table_with_id_pk_sql_with_query(
    evr_id_pk_basic_sql,
):
    rendered_value = get_renderer_impl(
        object_name=evr_id_pk_basic_sql.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_id_pk_basic_sql)
    assert rendered_value[0].to_json_dict() == {
        "content_block_type": "table",
        "header_row": ["Unexpected Value", "Count", "pk_1", "pk_2"],
        "styling": {"body": {"classes": ["table-bordered", "table-sm", "mt-3"]}},
        "table": [
            ["giraffe", 1, "3", "three"],
            ["lion", 1, "4", "four"],
            ["zebra", 1, "5", "five"],
        ],
    }
    assert rendered_value[1].to_json_dict() == {
        "collapse": [
            {
                "content_block_type": "string_template",
                "string_template": {
                    "tag": "code",
                    "template": "SELECT animals, pk_1, pk_2 \n"
                    "FROM animal_names \n"
                    "WHERE animals IS NOT NULL AND "
                    "(animals NOT IN ('cat', "
                    "'fish', 'dog'))",
                },
            }
        ],
        "collapse_toggle_link": "To retrieve all unexpected values...",
        "content_block_type": "collapse",
        "inline_link": False,
    }


def test_ValidationResultsTableContentBlockRenderer_get_unexpected_table_with_id_pk_sql_with_query_with_sampled_table(
    evr_id_pk_basic_sql,
):
    new_index = [
        {"animals": "giraffe", "pk_1": 3},
        {"animals": "lion", "pk_1": 4},
        {"animals": "zebra", "pk_1": 5},
        {"animals": "zebra", "pk_1": 6},
        {"animals": "zebra", "pk_1": 7},
        {"animals": "zebra", "pk_1": 8},
        {"animals": "zebra", "pk_1": 9},
        {"animals": "zebra", "pk_1": 10},
        {"animals": "zebra", "pk_1": 11},
        {"animals": "zebra", "pk_1": 12},
        {"animals": "zebra", "pk_1": 13},
        {"animals": "zebra", "pk_1": 14},
        {"animals": "zebra", "pk_1": 15},
    ]
    new_unexpected_list = [
        "giraffe",
        "lion",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
        "zebra",
    ]
    evr_id_pk_basic_sql.result["unexpected_index_column_names"] = ["pk_1"]
    evr_id_pk_basic_sql.result["partial_unexpected_index_list"] = new_index
    evr_id_pk_basic_sql.result["partial_unexpected_list"] = new_unexpected_list
    evr_id_pk_basic_sql.result["unexpected_count"] = 100
    evr_id_pk_basic_sql.result["unexpected_index_list"] = new_index
    evr_id_pk_basic_sql.result["unexpected_list"] = new_unexpected_list

    rendered_value = get_renderer_impl(
        object_name=evr_id_pk_basic_sql.expectation_config.expectation_type,
        renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
    )[1](result=evr_id_pk_basic_sql)
    assert rendered_value[0].to_json_dict() == {
        "content_block_type": "table",
        "header_row": ["Sampled Unexpected Values", "Count", "pk_1"],
        "styling": {"body": {"classes": ["table-bordered", "table-sm", "mt-3"]}},
        "table": [
            ["giraffe", 1, "3"],
            ["lion", 1, "4"],
            ["zebra", 11, "5, 6, 7, 8, 9, 10, 11, 12, 13, 14, ..."],
        ],
    }



import logging
import traceback
from typing import Any, Callable, Dict, Optional, Type, Union

from great_expectations.alias_types import JSONValues  # noqa: TCH001
from great_expectations.core import (
    ExpectationConfiguration,
    ExpectationValidationResult,
)
from great_expectations.expectations.registry import (
    _registered_renderers,
    get_renderer_impl,
)
from great_expectations.render import (
    CollapseContent,
    LegacyRendererType,
    RenderedComponentContent,
    RenderedMarkdownContent,
    RenderedStringTemplateContent,
    TextContent,
)
from great_expectations.render.renderer.renderer import Renderer

logger = logging.getLogger(__name__)


class ContentBlockRenderer(Renderer):
    _rendered_component_type: Type[RenderedComponentContent] = TextContent
    _default_header = ""

    _default_content_block_styling: Dict[str, JSONValues] = {"classes": ["col-12"]}

    _default_element_styling = {}

    @classmethod
    def validate_input(cls, render_object: Any) -> None:
        pass

    @classmethod
    def render(
        cls, render_object: Any, **kwargs
    ) -> Union[_rendered_component_type, Any, None]:
        cls.validate_input(render_object)
        exception_list_content_block: bool = kwargs.get(
            "exception_list_content_block", False
        )

        data_docs_exception_message = """\
An unexpected Exception occurred during data docs rendering.  Because of this error, certain parts of data docs will \
not be rendered properly and/or may not appear altogether.  Please use the trace, included in this message, to \
diagnose and repair the underlying issue.  Detailed information follows:
        """

        runtime_configuration = {
            "styling": cls._get_element_styling(),
            "include_column_name": kwargs.pop("include_column_name", None),
        }

        # The specific way we render the render_object is contingent on the type of the object
        render_fn: Callable
        if isinstance(render_object, list):
            render_fn = cls._render_list
        else:
            render_fn = cls._render_other

        result = render_fn(
            render_object,
            exception_list_content_block,
            runtime_configuration,
            data_docs_exception_message,
            kwargs,
        )
        return result

    @classmethod
    def _render_list(
        cls,
        render_object: list,
        exception_list_content_block: bool,
        runtime_configuration: dict,
        data_docs_exception_message: str,
        kwargs: dict,
    ) -> Optional[_rendered_component_type]:
        """Helper method to render list render_objects - refer to `render` for more context"""
        blocks = []
        has_failed_evr = (
            False if isinstance(render_object[0], ExpectationValidationResult) else None
        )
        for obj_ in render_object:

            expectation_type = cls._get_expectation_type(obj_)

            content_block_fn = cls._get_content_block_fn(expectation_type)

            if isinstance(obj_, ExpectationValidationResult) and not obj_.success:
                has_failed_evr = True

            if content_block_fn is not None and not exception_list_content_block:
                try:
                    if isinstance(obj_, ExpectationValidationResult):
                        expectation_config = obj_.expectation_config
                        result = content_block_fn(
                            configuration=expectation_config,
                            result=obj_,
                            runtime_configuration=runtime_configuration,
                            **kwargs,
                        )
                    else:
                        result = content_block_fn(
                            configuration=obj_,
                            runtime_configuration=runtime_configuration,
                            **kwargs,
                        )
                except Exception as e:
                    exception_traceback = traceback.format_exc()
                    exception_message = (
                        data_docs_exception_message
                        + f'{type(e).__name__}: "{str(e)}".  Traceback: "{exception_traceback}".'
                    )
                    logger.error(exception_message)

                    if isinstance(obj_, ExpectationValidationResult):
                        content_block_fn = cls._get_content_block_fn(
                            "_missing_content_block_fn"
                        )
                        expectation_config = obj_.expectation_config
                        result = content_block_fn(
                            configuration=expectation_config,
                            result=obj_,
                            runtime_configuration=runtime_configuration,
                            **kwargs,
                        )
                    else:
                        content_block_fn = cls._missing_content_block_fn
                        result = content_block_fn(
                            configuration=obj_,
                            runtime_configuration=runtime_configuration,
                            **kwargs,
                        )
            else:
                if isinstance(obj_, ExpectationValidationResult):
                    content_block_fn = (
                        cls._missing_content_block_fn
                        if exception_list_content_block
                        else cls._get_content_block_fn("_missing_content_block_fn")
                    )
                    expectation_config = obj_.expectation_config
                    result = content_block_fn(
                        configuration=expectation_config,
                        result=obj_,
                        runtime_configuration=runtime_configuration,
                        **kwargs,
                    )
                else:
                    content_block_fn = cls._missing_content_block_fn
                    result = content_block_fn(
                        configuration=obj_,
                        runtime_configuration=runtime_configuration,
                        **kwargs,
                    )

            if result is not None:
                if isinstance(obj_, ExpectationConfiguration):
                    expectation_meta_notes = cls._render_expectation_meta_notes(obj_)
                    if expectation_meta_notes:
                        # this adds collapse content block to expectation string
                        result[0] = [result[0], expectation_meta_notes]

                    horizontal_rule = RenderedStringTemplateContent(
                        **{
                            "content_block_type": "string_template",
                            "string_template": {
                                "template": "",
                                "tag": "hr",
                                "styling": {
                                    "classes": ["mt-1", "mb-1"],
                                },
                            },
                            "styling": {
                                "parent": {"styles": {"list-style-type": "none"}}
                            },
                        }
                    )
                    result.append(horizontal_rule)

                blocks += result

        if len(blocks) > 0:
            rendered_component_type_init_kwargs = {
                cls._content_block_type: blocks,
                "styling": cls._get_content_block_styling(),
            }
            rendered_component_type_default_init_kwargs = getattr(
                cls, "_rendered_component_default_init_kwargs", {}
            )
            rendered_component_type_init_kwargs.update(
                rendered_component_type_default_init_kwargs
            )
            content_block = cls._rendered_component_type(
                **rendered_component_type_init_kwargs
            )
            cls._process_content_block(
                content_block,
                has_failed_evr=has_failed_evr,
                render_object=render_object,
            )

            return content_block
        else:
            return None

    @classmethod
    def _render_other(
        cls,
        render_object: Any,
        exception_list_content_block: bool,
        runtime_configuration: dict,
        data_docs_exception_message: str,
        kwargs: dict,
    ) -> Any:
        """Helper method to render non-list render_objects - refer to `render` for more context"""
        expectation_type = cls._get_expectation_type(render_object)

        content_block_fn = get_renderer_impl(
            object_name=expectation_type, renderer_type=LegacyRendererType.PRESCRIPTIVE
        )
        content_block_fn = content_block_fn[1] if content_block_fn else None
        if content_block_fn is not None and not exception_list_content_block:
            try:
                if isinstance(render_object, ExpectationValidationResult):
                    result = content_block_fn(
                        result=render_object,
                        runtime_configuration=runtime_configuration,
                        **kwargs,
                    )
                else:
                    result = content_block_fn(
                        configuration=render_object,
                        runtime_configuration=runtime_configuration,
                        **kwargs,
                    )
            except Exception as e:
                exception_traceback = traceback.format_exc()
                exception_message = (
                    data_docs_exception_message
                    + f'{type(e).__name__}: "{str(e)}".  Traceback: "{exception_traceback}".'
                )
                logger.error(exception_message)

                if isinstance(render_object, ExpectationValidationResult):
                    content_block_fn = cls._get_content_block_fn(
                        "_missing_content_block_fn"
                    )
                    result = content_block_fn(
                        result=render_object,
                        runtime_configuration=runtime_configuration,
                        **kwargs,
                    )
                else:
                    content_block_fn = cls._missing_content_block_fn
                    result = content_block_fn(
                        configuration=render_object,
                        runtime_configuration=runtime_configuration,
                        **kwargs,
                    )
        else:
            if isinstance(render_object, ExpectationValidationResult):
                content_block_fn = (
                    cls._missing_content_block_fn
                    if exception_list_content_block
                    else cls._get_content_block_fn("_missing_content_block_fn")
                )
                result = content_block_fn(
                    result=render_object,
                    runtime_configuration=runtime_configuration,
                    **kwargs,
                )
            else:
                content_block_fn = cls._missing_content_block_fn
                result = content_block_fn(
                    configuration=render_object,
                    runtime_configuration=runtime_configuration,
                    **kwargs,
                )
        if result is not None:
            if isinstance(render_object, ExpectationConfiguration):
                expectation_meta_notes = cls._render_expectation_meta_notes(
                    render_object
                )
                if expectation_meta_notes:
                    result.append(expectation_meta_notes)
        return result

    @classmethod
    def _render_expectation_meta_notes(cls, expectation):
        if not expectation.meta.get("notes"):
            return None
        else:
            collapse_link = RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": ""},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": ["fas", "fa-comment", "text-info"],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                }
            )
            notes = expectation.meta["notes"]
            note_content = None

            if isinstance(notes, str):
                note_content = [notes]

            elif isinstance(notes, list):
                note_content = notes

            elif isinstance(notes, dict):
                if "format" in notes:
                    if notes["format"] == "string":
                        if isinstance(notes["content"], str):
                            note_content = [notes["content"]]
                        elif isinstance(notes["content"], list):
                            note_content = notes["content"]
                        else:
                            logger.warning(
                                "Unrecognized Expectation suite notes format. Skipping rendering."
                            )

                    elif notes["format"] == "markdown":
                        if isinstance(notes["content"], str):
                            note_content = [
                                RenderedMarkdownContent(
                                    **{
                                        "content_block_type": "markdown",
                                        "markdown": notes["content"],
                                        "styling": {
                                            "parent": {"styles": {"color": "red"}}
                                        },
                                    }
                                )
                            ]
                        elif isinstance(notes["content"], list):
                            note_content = [
                                RenderedMarkdownContent(
                                    **{
                                        "content_block_type": "markdown",
                                        "markdown": note,
                                        "styling": {"parent": {}},
                                    }
                                )
                                for note in notes["content"]
                            ]
                        else:
                            logger.warning(
                                "Unrecognized Expectation suite notes format. Skipping rendering."
                            )
                else:
                    logger.warning(
                        "Unrecognized Expectation suite notes format. Skipping rendering."
                    )

            notes_block = TextContent(
                **{
                    "content_block_type": "text",
                    "subheader": "Notes:",
                    "text": note_content,
                    "styling": {
                        "classes": ["col-12", "mt-2", "mb-2"],
                        "parent": {"styles": {"list-style-type": "none"}},
                    },
                }
            )

            return CollapseContent(
                **{
                    "collapse_toggle_link": collapse_link,
                    "collapse": [notes_block],
                    "inline_link": True,
                    "styling": {
                        "body": {"classes": ["card", "card-body", "p-1"]},
                        "parent": {"styles": {"list-style-type": "none"}},
                    },
                }
            )

    @classmethod
    def _process_content_block(
        cls, content_block, has_failed_evr, render_object=None
    ) -> None:
        header = cls._get_header()
        if header != "":
            content_block.header = header

    @classmethod
    def _get_content_block_fn(cls, expectation_type):
        content_block_fn = get_renderer_impl(
            object_name=expectation_type, renderer_type=LegacyRendererType.PRESCRIPTIVE
        )
        return content_block_fn[1] if content_block_fn else None

    @classmethod
    def list_available_expectations(cls):
        expectations = [
            object_name
            for object_name in _registered_renderers
            if object_name.startswith("expect_")
        ]
        return expectations

    @classmethod
    def _missing_content_block_fn(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        return []

    @classmethod
    def _get_content_block_styling(cls):
        return cls._default_content_block_styling

    @classmethod
    def _get_element_styling(cls):
        return cls._default_element_styling

    @classmethod
    def _get_header(cls):
        return cls._default_header



import datetime
import os

import pytest
from google.cloud import bigquery

import great_expectations as gx
from great_expectations import DataContext
from great_expectations.core.batch import BatchRequest, RuntimeBatchRequest
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.validator.validator import Validator

"""
What does this test and why?

This integration test is part of the deprecation of the bigquery_temp_table parameter. As of GX 0.15.3,
tables that are created as the result of a query are created as permanent tables with an expiration of 24 hours, with
more information to be found:

- https://github.com/great-expectations/great_expectations/pull/4925
- https://stackoverflow.com/questions/20673986/how-to-create-temporary-table-in-google-bigquery

This integration test tests the following:

1. get_validator() will create a temp_table with the expected expiration
2. passing in `bigquery_temp_table` as part of `batch_spec_passthrough` will raise a DeprecationWarning.
3. the validator.head method (and corresponding metric) will work on bigquery datasources even when temp table creation
    is turned off
"""

gcp_project: str = os.environ.get("GE_TEST_GCP_PROJECT")
if not gcp_project:
    raise ValueError(
        "Environment Variable GE_TEST_GCP_PROJECT is required to run BigQuery integration tests"
    )
bigquery_dataset: str = "test_ci"
CONNECTION_STRING: str = f"bigquery://{gcp_project}/{bigquery_dataset}"

yaml = YAMLHandler()

context: DataContext = gx.get_context()

datasource_yaml: str = f"""
name: my_bigquery_datasource
class_name: Datasource
execution_engine:
  class_name: SqlAlchemyExecutionEngine
  connection_string: bigquery://<GCP_PROJECT_NAME>/<BIGQUERY_DATASET>
data_connectors:
   default_runtime_data_connector_name:
       class_name: RuntimeDataConnector
       assets:
            asset_a:
                batch_identifiers:
                    - default_identifier_name
   default_inferred_data_connector_name:
       class_name: InferredAssetSqlDataConnector
       include_schema_name: true
"""

datasource_yaml: str = datasource_yaml.replace(
    "bigquery://<GCP_PROJECT_NAME>/<BIGQUERY_DATASET>",
    CONNECTION_STRING,
)

context.add_datasource(**yaml.load(datasource_yaml))

batch_request: RuntimeBatchRequest = RuntimeBatchRequest(
    datasource_name="my_bigquery_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="asset_a",  # this can be anything that identifies this data
    runtime_parameters={"query": "SELECT * from demo.taxi_data LIMIT 10"},
    batch_identifiers={"default_identifier_name": "default_identifier"},
)
context.add_expectation_suite(expectation_suite_name="test_suite")
validator: Validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
assert validator

# What is the name of the temp_table that was created as part of `get_validator`?
temp_table_name: str = validator.active_batch.data.selectable.description
client: bigquery.Client = bigquery.Client()
project: str = client.project
dataset_ref: bigquery.DatasetReference = bigquery.DatasetReference(
    project, bigquery_dataset
)
table_ref: bigquery.TableReference = dataset_ref.table(temp_table_name)
table: bigquery.Table = client.get_table(table_ref)

# what is its expiration
expected_expiration: datetime.datetime = datetime.datetime.now(
    datetime.timezone.utc
) + datetime.timedelta(days=1)
assert table.expires <= expected_expiration

# Ensure that passing in `bigquery_temp_table` will fire off a DeprecationWarning
batch_request: RuntimeBatchRequest = RuntimeBatchRequest(
    datasource_name="my_bigquery_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="asset_a",  # this can be anything that identifies this data
    runtime_parameters={"query": "SELECT * from demo.taxi_data LIMIT 10"},
    batch_identifiers={"default_identifier_name": "default_identifier"},
    batch_spec_passthrough={
        "bigquery_temp_table": temp_table_name
    },  # this is the name of the table you would like to use a 'temp_table'
)

context.add_expectation_suite(expectation_suite_name="test_suite_2")
with pytest.warns(DeprecationWarning):
    validator: Validator = context.get_validator(
        batch_request=batch_request, expectation_suite_name="test_suite_2"
    )

# Testing `validator.head()` when using BigQuery without temp tables
batch_request_without_temp_table: RuntimeBatchRequest = RuntimeBatchRequest(
    datasource_name="my_bigquery_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="asset_a",  # this can be anything that identifies this data
    runtime_parameters={"query": "SELECT * from demo.taxi_data LIMIT 10"},
    batch_identifiers={"default_identifier_name": "default_identifier"},
    batch_spec_passthrough={"create_temp_table": False},
)
validator_without_temp_table = context.get_validator(
    batch_request=batch_request_without_temp_table,
    expectation_suite_name="test_suite_2",
)

df = validator_without_temp_table.head(n_rows=5, fetch_all=False)
assert not df.empty



from typing import Any, Dict, List

import pandas as pd

from great_expectations.checkpoint.types.checkpoint_result import CheckpointResult
from great_expectations.core import (
    ExpectationConfiguration,
    ExpectationSuiteValidationResult,
    IDDict,
)
from great_expectations.core.batch import Batch, BatchDefinition
from great_expectations.data_context.data_context.base_data_context import (
    BaseDataContext,
)
from great_expectations.data_context.types.base import (
    CheckpointConfig,
    DataContextConfig,
    InMemoryStoreBackendDefaults,
)
from great_expectations.execution_engine import PandasExecutionEngine
from great_expectations.util import filter_properties_dict
from great_expectations.validator.validator import Validator

# Snippet: example data frame for result_format
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/pandas_df_for_result_format">
dataframe = pd.DataFrame(
    {
        "pk_column": ["zero", "one", "two", "three", "four", "five", "six", "seven"],
        "my_var": ["A", "B", "B", "C", "C", "C", "D", "D"],
        "my_numbers": [1.0, 2.0, 2.0, 3.0, 3.0, 3.0, 4.0, 4.0],
    }
)
# </snippet>


# NOTE: The following code is only for testing and can be ignored by users.
data_context_config: DataContextConfig = DataContextConfig(
    datasources={  # type: ignore[arg-type]
        "pandas_datasource": {
            "execution_engine": {
                "class_name": "PandasExecutionEngine",
                "module_name": "great_expectations.execution_engine",
            },
            "class_name": "Datasource",
            "module_name": "great_expectations.datasource",
            "data_connectors": {
                "runtime_data_connector": {
                    "class_name": "RuntimeDataConnector",
                    "batch_identifiers": [
                        "id_key_0",
                        "id_key_1",
                    ],
                }
            },
        },
    },
    expectations_store_name="expectations_store",
    validations_store_name="validations_store",
    evaluation_parameter_store_name="evaluation_parameter_store",
    checkpoint_store_name="checkpoint_store",
    store_backend_defaults=InMemoryStoreBackendDefaults(),
)
context = BaseDataContext(project_config=data_context_config)
batch_definition = BatchDefinition(
    datasource_name="pandas_datasource",
    data_connector_name="runtime_data_connector",
    data_asset_name="my_asset",
    batch_identifiers=IDDict({}),
    batch_spec_passthrough=None,
)
batch = Batch(
    data=dataframe,
    batch_definition=batch_definition,
)
engine = PandasExecutionEngine()
my_validator: Validator = Validator(
    execution_engine=engine,
    data_context=context,
    batches=[
        batch,
    ],
)

# Expectation-level Configuration
# Snippet: result_format BOOLEAN example
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_boolean_example">
validation_result = my_validator.expect_column_values_to_be_in_set(
    column="my_var",
    value_set=["A", "B"],
    result_format={"result_format": "BOOLEAN_ONLY"},
)
# </snippet>

# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_boolean_example_output">
assert validation_result.success == False
assert validation_result.result == {}
# </snippet>


# Snippet: result_format BASIC example with set
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_basic_example_set">
validation_result = my_validator.expect_column_values_to_be_in_set(
    column="my_var", value_set=["A", "B"], result_format={"result_format": "BASIC"}
)
# </snippet>
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_basic_example_set_output">
assert validation_result.success == False
assert validation_result.result == {
    "element_count": 8,
    "unexpected_count": 5,
    "unexpected_percent": 62.5,
    "partial_unexpected_list": ["C", "C", "C", "D", "D"],
    "missing_count": 0,
    "missing_percent": 0.0,
    "unexpected_percent_total": 62.5,
    "unexpected_percent_nonmissing": 62.5,
}
# </snippet>

# Snippet: result_format BASIC example with aggregate
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_basic_example_agg">
validation_result = my_validator.expect_column_mean_to_be_between(
    column="my_numbers", min_value=0.0, max_value=10.0, result_format="BASIC"
)
# </snippet>
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_basic_example_agg_output">
assert validation_result.success == True
assert validation_result.result == {"observed_value": 2.75}
# </snippet>


# Snippet: result_format SUMMARY example with set
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_summary_example_set">
validation_result = my_validator.expect_column_values_to_be_in_set(
    column="my_var",
    value_set=["A", "B"],
    result_format={
        "result_format": "SUMMARY",
        "unexpected_index_column_names": ["pk_column"],
        "return_unexpected_index_query": True,
    },
)
# </snippet>

# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_summary_example_set_output">
assert validation_result.success == False
assert validation_result.result == {
    "element_count": 8,
    "unexpected_count": 5,
    "unexpected_percent": 62.5,
    "partial_unexpected_list": ["C", "C", "C", "D", "D"],
    "unexpected_index_column_names": ["pk_column"],
    "missing_count": 0,
    "missing_percent": 0.0,
    "unexpected_percent_total": 62.5,
    "unexpected_percent_nonmissing": 62.5,
    "partial_unexpected_index_list": [
        {"my_var": "C", "pk_column": "three"},
        {"my_var": "C", "pk_column": "four"},
        {"my_var": "C", "pk_column": "five"},
        {"my_var": "D", "pk_column": "six"},
        {"my_var": "D", "pk_column": "seven"},
    ],
    "partial_unexpected_counts": [
        {"value": "C", "count": 3},
        {"value": "D", "count": 2},
    ],
}
# </snippet>

# Snippet: result_format SUMMARY example with agg
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_summary_example_agg">
validation_result = my_validator.expect_column_mean_to_be_between(
    column="my_numbers", min_value=0.0, max_value=10.0, result_format="SUMMARY"
)
# </snippet>

# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_summary_example_agg_output">
assert validation_result.success == True
assert validation_result.result == {"observed_value": 2.75}
# </snippet>

# Snippet: result_format COMPLETE example with set
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_complete_example_set">
validation_result = my_validator.expect_column_values_to_be_in_set(
    column="my_var",
    value_set=["A", "B"],
    result_format={
        "result_format": "COMPLETE",
        "unexpected_index_column_names": ["pk_column"],
        "return_unexpected_index_query": True,
    },
)
# </snippet>

# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_complete_example_set_output">
assert validation_result.success == False
assert validation_result.result == {
    "element_count": 8,
    "unexpected_count": 5,
    "unexpected_percent": 62.5,
    "partial_unexpected_list": ["C", "C", "C", "D", "D"],
    "unexpected_index_column_names": ["pk_column"],
    "missing_count": 0,
    "missing_percent": 0.0,
    "unexpected_percent_total": 62.5,
    "unexpected_percent_nonmissing": 62.5,
    "partial_unexpected_index_list": [
        {"my_var": "C", "pk_column": "three"},
        {"my_var": "C", "pk_column": "four"},
        {"my_var": "C", "pk_column": "five"},
        {"my_var": "D", "pk_column": "six"},
        {"my_var": "D", "pk_column": "seven"},
    ],
    "partial_unexpected_counts": [
        {"value": "C", "count": 3},
        {"value": "D", "count": 2},
    ],
    "unexpected_list": ["C", "C", "C", "D", "D"],
    "unexpected_index_list": [
        {"my_var": "C", "pk_column": "three"},
        {"my_var": "C", "pk_column": "four"},
        {"my_var": "C", "pk_column": "five"},
        {"my_var": "D", "pk_column": "six"},
        {"my_var": "D", "pk_column": "seven"},
    ],
    "unexpected_index_query": [3, 4, 5, 6, 7],
}
# </snippet>

# Snippet: result_format COMPLETE example with agg
# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_complete_example_agg">
validation_result = my_validator.expect_column_mean_to_be_between(
    column="my_numbers", min_value=0.0, max_value=10.0, result_format="COMPLETE"
)
# </snippet>

# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_complete_example_agg_output">
assert validation_result.success == True
assert validation_result.result == {"observed_value": 2.75}
# </snippet>


# Checkpoint
# NOTE: The following code is only for testing and can be ignored by users.
context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
test_suite = context.get_expectation_suite(expectation_suite_name="test_suite")

expectation_config = ExpectationConfiguration(
    expectation_type="expect_column_values_to_be_in_set",
    kwargs={
        "column": "my_var",
        "value_set": ["A", "B"],
    },
)
test_suite.add_expectation(expectation_configuration=expectation_config)

test_suite.expectation_suite_name = "test_suite"
context.add_or_update_expectation_suite(
    expectation_suite=test_suite,
)

# <snippet name="tests/integration/docusaurus/reference/core_concepts/result_format/result_format_checkpoint_example">
checkpoint_dict: dict = {
    "name": "my_checkpoint",
    "config_version": 1.0,
    "class_name": "Checkpoint",  # or SimpleCheckpoint
    "module_name": "great_expectations.checkpoint",
    "template_name": None,
    "run_name_template": "%Y-%M-foo-bar-template-test",
    "expectation_suite_name": None,
    "batch_request": None,
    "profilers": [],
    "action_list": [
        {
            "name": "store_validation_result",
            "action": {"class_name": "StoreValidationResultAction"},
        },
        {
            "name": "store_evaluation_params",
            "action": {"class_name": "StoreEvaluationParametersAction"},
        },
        {
            "name": "update_data_docs",
            "action": {"class_name": "UpdateDataDocsAction"},
        },
    ],
    "validations": [],
    "runtime_configuration": {
        "result_format": {
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_column"],
            "return_unexpected_index_query": True,
        },
    },
}
# </snippet>
batch_request = {
    "datasource_name": "pandas_datasource",
    "data_connector_name": "runtime_data_connector",
    "data_asset_name": "IN_MEMORY_DATA_ASSET",
    "runtime_parameters": {
        "batch_data": dataframe,
    },
    "batch_identifiers": {
        "id_key_0": 1234567890,
    },
}

checkpoint_config = CheckpointConfig(**checkpoint_dict)
context.add_or_update_checkpoint(
    **filter_properties_dict(
        properties=checkpoint_config.to_json_dict(),
        clean_falsy=True,
    ),
)
context._save_project_config()

result: CheckpointResult = context.run_checkpoint(
    checkpoint_name="my_checkpoint",
    expectation_suite_name="test_suite",
    batch_request=batch_request,
)
evrs: List[ExpectationSuiteValidationResult] = result.list_validation_results()

result_index_list: List[Dict[str, Any]] = evrs[0]["results"][0]["result"][
    "unexpected_index_list"
]
assert result_index_list == [
    {"my_var": "C", "pk_column": "three"},
    {"my_var": "C", "pk_column": "four"},
    {"my_var": "C", "pk_column": "five"},
    {"my_var": "D", "pk_column": "six"},
    {"my_var": "D", "pk_column": "seven"},
]

result_index_query: List[int] = evrs[0]["results"][0]["result"][
    "unexpected_index_query"
]
assert result_index_query == [3, 4, 5, 6, 7]
partial_unexpected_counts: List[Dict[str, Any]] = evrs[0]["results"][0]["result"][
    "partial_unexpected_counts"
]
assert partial_unexpected_counts == [
    {"value": "C", "count": 3},
    {"value": "D", "count": 2},
]



from __future__ import annotations

import abc
import enum
from dataclasses import dataclass
from typing import cast

from great_expectations.core.util import convert_to_json_serializable
from great_expectations.exceptions import ProfilerConfigurationError
from great_expectations.types import SerializableDictDot


@dataclass(frozen=True)
class CardinalityLimit(abc.ABC, SerializableDictDot):
    name: str


@dataclass(frozen=True)
class RelativeCardinalityLimit(CardinalityLimit):
    max_proportion_unique: float
    metric_name_defining_limit: str = "column.unique_proportion"

    def to_json_dict(self) -> dict:
        return convert_to_json_serializable(
            {
                "name": self.name,
                "max_proportion_unique": self.max_proportion_unique,
                "metric_name_defining_limit": self.metric_name_defining_limit,
            }
        )


@dataclass(frozen=True)
class AbsoluteCardinalityLimit(CardinalityLimit):
    max_unique_values: int
    metric_name_defining_limit: str = "column.distinct_values.count"

    def to_json_dict(self) -> dict:
        return convert_to_json_serializable(
            {
                "name": self.name,
                "max_proportion_unique": self.max_unique_values,
                "metric_name_defining_limit": self.metric_name_defining_limit,
            }
        )


class CardinalityLimitMode(enum.Enum):
    """Preset limits based on unique values (cardinality)

    Defines relative (ratio) and absolute number of records (table rows) that
    correspond to each cardinality category.

    Used to determine appropriate Expectation configurations based on data.
    """

    ZERO = AbsoluteCardinalityLimit("ZERO", 0)
    ONE = AbsoluteCardinalityLimit("ONE", 1)
    TWO = AbsoluteCardinalityLimit("TWO", 2)
    VERY_FEW = AbsoluteCardinalityLimit("VERY_FEW", 10)
    FEW = AbsoluteCardinalityLimit("FEW", 100)
    SOME = AbsoluteCardinalityLimit("SOME", 1000)
    MANY = AbsoluteCardinalityLimit("MANY", 10000)
    VERY_MANY = AbsoluteCardinalityLimit("VERY_MANY", 100000)
    UNIQUE = RelativeCardinalityLimit("UNIQUE", 1.0)
    ABS_10 = AbsoluteCardinalityLimit("ABS_10", 10)
    ABS_100 = AbsoluteCardinalityLimit("ABS_100", 100)
    ABS_1000 = AbsoluteCardinalityLimit("ABS_1000", 1000)
    ABS_10_000 = AbsoluteCardinalityLimit("ABS_10_000", int(1e4))
    ABS_100_000 = AbsoluteCardinalityLimit("ABS_100_000", int(1e5))
    ABS_1_000_000 = AbsoluteCardinalityLimit("ABS_1_000_000", int(1e6))
    ABS_10_000_000 = AbsoluteCardinalityLimit("ABS_10_000_000", int(1e7))
    ABS_100_000_000 = AbsoluteCardinalityLimit("ABS_100_000_000", int(1e8))
    ABS_1_000_000_000 = AbsoluteCardinalityLimit("ABS_1_000_000_000", int(1e9))
    REL_0 = RelativeCardinalityLimit("REL_0", 0.0)
    REL_001 = RelativeCardinalityLimit("REL_001", 1e-5)
    REL_01 = RelativeCardinalityLimit("REL_01", 1e-4)
    REL_0_1 = RelativeCardinalityLimit("REL_0_1", 1e-3)
    REL_1 = RelativeCardinalityLimit("REL_1", 1e-2)
    REL_10 = RelativeCardinalityLimit("REL_10", 0.10)
    REL_25 = RelativeCardinalityLimit("REL_25", 0.25)
    REL_50 = RelativeCardinalityLimit("REL_50", 0.50)
    REL_75 = RelativeCardinalityLimit("REL_75", 0.75)
    REL_100 = RelativeCardinalityLimit("REL_100", 1.0)
    ONE_PCT = RelativeCardinalityLimit("ONE_PCT", 0.01)
    TEN_PCT = RelativeCardinalityLimit("TEN_PCT", 0.10)


class CardinalityChecker:
    """Handles cardinality checking given cardinality limit mode and measured value.

    This class also validates cardinality limit settings and converts from
    various types of settings. You can choose one of the attributes listed
    below to create an instance.

    Attributes:
        cardinality_limit_mode: CardinalityLimitMode or string name of the mode
            defining the maximum allowable cardinality.
        max_unique_values: number of max unique rows for a custom
            cardinality limit.
        max_proportion_unique: proportion of unique values for a
            custom cardinality limit.
    """

    SUPPORTED_CARDINALITY_LIMIT_MODE_CLASSES = (
        AbsoluteCardinalityLimit,
        RelativeCardinalityLimit,
    )
    SUPPORTED_LIMIT_MODE_CLASS_NAMES = (
        mode.__name__ for mode in SUPPORTED_CARDINALITY_LIMIT_MODE_CLASSES
    )
    SUPPORTED_CARDINALITY_LIMIT_MODE_STRINGS = (
        mode.name for mode in CardinalityLimitMode
    )

    def __init__(
        self,
        cardinality_limit_mode: str | CardinalityLimitMode | dict | None = None,
        max_unique_values: int | None = None,
        max_proportion_unique: float | None = None,
    ) -> None:
        self._cardinality_limit_mode = self._convert_to_cardinality_limit_mode(
            cardinality_limit_mode=cardinality_limit_mode,
            max_unique_values=max_unique_values,
            max_proportion_unique=max_proportion_unique,
        )

    @property
    def cardinality_limit_mode(
        self,
    ) -> AbsoluteCardinalityLimit | RelativeCardinalityLimit:
        return self._cardinality_limit_mode

    def cardinality_within_limit(self, metric_value: float) -> bool:
        """Determine if the cardinality is within configured limit.

        The metric_value supplied should be either a proportion of unique values
        or number of unique values based on the configured cardinality limit.

        Args:
            metric_value: int if number of unique values, float if proportion
                of unique values.

        Returns:
            Boolean of whether the cardinality is within the configured limit
        """
        self._validate_metric_value(metric_value=metric_value)
        if isinstance(self._cardinality_limit_mode, AbsoluteCardinalityLimit):
            return metric_value <= self._cardinality_limit_mode.max_unique_values

        if isinstance(self._cardinality_limit_mode, RelativeCardinalityLimit):
            return metric_value <= self._cardinality_limit_mode.max_proportion_unique

        raise ValueError(
            f'Unknown "cardinality_limit_mode" mode "{self._cardinality_limit_mode}" encountered.'
        )

    @staticmethod
    def _validate_metric_value(metric_value: float) -> None:
        if not isinstance(metric_value, (int, float)):
            raise ProfilerConfigurationError(
                f"Value of measured cardinality must be of type int or float, you provided {type(metric_value)}"
            )

        if metric_value < 0.00:
            raise ProfilerConfigurationError(
                f"Value of cardinality (number of rows or percent unique) should be greater than 0.00, your value is {metric_value}"
            )

    @staticmethod
    def _to_cardinality_limit_mode(
        cardinality_limit_mode: str | CardinalityLimitMode | dict | None = None,
    ):
        if isinstance(cardinality_limit_mode, str):
            try:
                return CardinalityLimitMode[cardinality_limit_mode.upper()].value
            except KeyError:
                raise ProfilerConfigurationError(
                    f"Please specify a supported cardinality mode. Supported cardinality modes are {[member.name for member in CardinalityLimitMode]}"
                )
        elif isinstance(cardinality_limit_mode, dict):
            validate_input_parameters(
                cardinality_limit_mode=cardinality_limit_mode.get("name"),
                max_unique_values=cardinality_limit_mode.get("max_unique_values"),
                max_proportion_unique=cardinality_limit_mode.get(
                    "max_proportion_unique"
                ),
                required_num_supplied_params=2,
            )
            try:
                return AbsoluteCardinalityLimit(
                    name=cardinality_limit_mode["name"],
                    max_unique_values=cardinality_limit_mode["max_unique_values"],
                    metric_name_defining_limit=cardinality_limit_mode[
                        "metric_name_defining_limit"
                    ],
                )
            except (KeyError, ValueError):
                try:
                    return RelativeCardinalityLimit(
                        name=cardinality_limit_mode["name"],
                        max_proportion_unique=cardinality_limit_mode[
                            "max_proportion_unique"
                        ],
                        metric_name_defining_limit=cardinality_limit_mode[
                            "metric_name_defining_limit"
                        ],
                    )
                except (KeyError, ValueError):
                    raise ProfilerConfigurationError(
                        f"Please specify a supported cardinality mode.  Supported cardinality modes are {[member.name for member in CardinalityLimitMode]}"
                    )
        else:
            return cast(CardinalityLimitMode, cardinality_limit_mode).value

    @staticmethod
    def _convert_to_cardinality_limit_mode(
        cardinality_limit_mode: str | CardinalityLimitMode | dict | None = None,
        max_unique_values: int | None = None,
        max_proportion_unique: float | None = None,
    ) -> AbsoluteCardinalityLimit | RelativeCardinalityLimit:
        validate_input_parameters(
            cardinality_limit_mode=cardinality_limit_mode,
            max_unique_values=max_unique_values,
            max_proportion_unique=max_proportion_unique,
        )

        if cardinality_limit_mode is not None:
            return CardinalityChecker._to_cardinality_limit_mode(cardinality_limit_mode)
        elif max_unique_values is not None:
            return AbsoluteCardinalityLimit(
                name=f"CUSTOM_ABS_{max_unique_values}",
                max_unique_values=max_unique_values,
            )
        else:
            assert (
                max_proportion_unique is not None
            ), "Guaranteed to have `max_proportion_unique` due to prior call to `validate_input_parameters`"
            return RelativeCardinalityLimit(
                name=f"CUSTOM_REL_{max_proportion_unique}",
                max_proportion_unique=max_proportion_unique,
            )


def validate_input_parameters(
    cardinality_limit_mode: str | CardinalityLimitMode | dict | None = None,
    max_unique_values: int | None = None,
    max_proportion_unique: float | None = None,
    required_num_supplied_params: int = 1,
) -> None:
    num_supplied_params: int = sum(
        0 if param is None else 1
        for param in (
            cardinality_limit_mode,
            max_unique_values,
            max_proportion_unique,
        )
    )
    if num_supplied_params != required_num_supplied_params:
        raise ProfilerConfigurationError(
            f"Please pass ONE of the following parameters: cardinality_limit_mode, max_unique_values, max_proportion_unique, you passed {num_supplied_params} parameters."
        )

    if cardinality_limit_mode is not None:
        if not isinstance(cardinality_limit_mode, (str, CardinalityLimitMode, dict)):
            raise ProfilerConfigurationError(
                f"Please specify a supported cardinality limit type, supported classes are {','.join(CardinalityChecker.SUPPORTED_LIMIT_MODE_CLASS_NAMES)} and supported strings are {','.join(CardinalityChecker.SUPPORTED_CARDINALITY_LIMIT_MODE_STRINGS)}"
            )

        if required_num_supplied_params == 2:
            try:
                assert isinstance(cardinality_limit_mode, str)
                return CardinalityLimitMode[cardinality_limit_mode.upper()].value
            except KeyError:
                raise ProfilerConfigurationError(
                    f"Please specify a supported cardinality mode. Supported cardinality modes are {[member.name for member in CardinalityLimitMode]}"
                )

    if max_unique_values is not None:
        if not isinstance(max_unique_values, int):
            raise ProfilerConfigurationError(
                f"Please specify an int, you specified a {type(max_unique_values)}"
            )

    if max_proportion_unique is not None:
        if not isinstance(max_proportion_unique, (float, int)):
            raise ProfilerConfigurationError(
                f"Please specify a float or int, you specified a {type(max_proportion_unique)}"
            )



import os

import pytest

import great_expectations.exceptions as gx_exceptions
from great_expectations.data_context.util import (
    PasswordMasker,
    parse_substitution_variable,
)
from great_expectations.exceptions.exceptions import StoreConfigurationError
from great_expectations.types import safe_deep_copy
from great_expectations.util import load_class


def test_load_class_raises_error_when_module_not_found():
    with pytest.raises(gx_exceptions.PluginModuleNotFoundError):
        load_class("foo", "bar")


def test_load_class_raises_error_when_class_not_found():
    with pytest.raises(gx_exceptions.PluginClassNotFoundError):
        load_class("TotallyNotARealClass", "great_expectations.datasource")


def test_load_class_raises_error_when_class_name_is_None():
    with pytest.raises(TypeError):
        load_class(None, "great_expectations.datasource")


def test_load_class_raises_error_when_class_name_is_not_string():
    for bad_input in [1, 1.3, ["a"], {"foo": "bar"}]:
        with pytest.raises(TypeError):
            load_class(bad_input, "great_expectations.datasource")


def test_load_class_raises_error_when_module_name_is_None():
    with pytest.raises(TypeError):
        load_class("foo", None)


def test_load_class_raises_error_when_module_name_is_not_string():
    for bad_input in [1, 1.3, ["a"], {"foo": "bar"}]:
        with pytest.raises(TypeError):
            load_class(bad_input, "great_expectations.datasource")


@pytest.mark.filterwarnings(
    "ignore:SQLAlchemy is not installed*:UserWarning:great_expectations.data_context.util"
)
def test_password_masker_mask_db_url(monkeypatch, tmp_path):
    """
    What does this test and why?
    The PasswordMasker.mask_db_url() should mask passwords consistently in database urls. The output of mask_db_url should be the same whether user_urlparse is set to True or False.
    This test uses database url examples from
    https://docs.sqlalchemy.org/en/14/core/engines.html#database-urls
    """
    # PostgreSQL (if installed in test environment)
    # default
    db_hostname = os.getenv("GE_TEST_LOCAL_DB_HOSTNAME", "localhost")
    try:
        assert (
            PasswordMasker.mask_db_url(
                f"postgresql://scott:tiger@{db_hostname}:65432/mydatabase"
            )
            == f"postgresql://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/mydatabase"
        )
    except ModuleNotFoundError:
        pass
    assert (
        PasswordMasker.mask_db_url(
            f"postgresql://scott:tiger@{db_hostname}:65432/mydatabase",
            use_urlparse=True,
        )
        == f"postgresql://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/mydatabase"
    )
    # missing port number, using urlparse
    assert (
        PasswordMasker.mask_db_url(
            f"postgresql://scott:tiger@{db_hostname}/mydatabase", use_urlparse=True
        )
        == f"postgresql://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}/mydatabase"
    )

    # psycopg2 (if installed in test environment)
    try:
        assert (
            PasswordMasker.mask_db_url(
                f"postgresql+psycopg2://scott:tiger@{db_hostname}:65432/mydatabase"
            )
            == f"postgresql+psycopg2://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/mydatabase"
        )
    except ModuleNotFoundError:
        pass
    assert (
        PasswordMasker.mask_db_url(
            f"postgresql+psycopg2://scott:tiger@{db_hostname}:65432/mydatabase",
            use_urlparse=True,
        )
        == f"postgresql+psycopg2://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/mydatabase"
    )

    # pg8000 (if installed in test environment)
    try:
        assert (
            PasswordMasker.mask_db_url(
                f"postgresql+pg8000://scott:tiger@{db_hostname}:65432/mydatabase"
            )
            == f"postgresql+pg8000://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/mydatabase"
        )
    except ModuleNotFoundError:
        pass
    assert (
        PasswordMasker.mask_db_url(
            f"postgresql+pg8000://scott:tiger@{db_hostname}:65432/mydatabase",
            use_urlparse=True,
        )
        == f"postgresql+pg8000://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/mydatabase"
    )

    # MySQL
    # default (if installed in test environment)
    try:
        assert (
            PasswordMasker.mask_db_url(f"mysql://scott:tiger@{db_hostname}:65432/foo")
            == f"mysql://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/foo"
        )
    except ModuleNotFoundError:
        pass

    assert (
        PasswordMasker.mask_db_url(
            f"mysql://scott:tiger@{db_hostname}:65432/foo", use_urlparse=True
        )
        == f"mysql://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/foo"
    )

    # mysqlclient (a maintained fork of MySQL-Python) (if installed in test environment)
    try:
        assert (
            PasswordMasker.mask_db_url(
                f"mysql+mysqldb://scott:tiger@{db_hostname}:65432/foo"
            )
            == f"mysql+mysqldb://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/foo"
        )
    except ModuleNotFoundError:
        pass
    assert (
        PasswordMasker.mask_db_url(
            f"mysql+mysqldb://scott:tiger@{db_hostname}:65432/foo", use_urlparse=True
        )
        == f"mysql+mysqldb://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/foo"
    )

    # PyMySQL (if installed in test environment)
    try:
        assert (
            PasswordMasker.mask_db_url(
                f"mysql+pymysql://scott:tiger@{db_hostname}:65432/foo"
            )
            == f"mysql+pymysql://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/foo"
        )
    except ModuleNotFoundError:
        pass
    assert (
        PasswordMasker.mask_db_url(
            f"mysql+pymysql://scott:tiger@{db_hostname}:65432/foo", use_urlparse=True
        )
        == f"mysql+pymysql://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:65432/foo"
    )

    # Oracle (if installed in test environment)
    url_host = os.getenv("GE_TEST_LOCALHOST_URL", "127.0.0.1")
    try:
        assert (
            PasswordMasker.mask_db_url(f"oracle://scott:tiger@{url_host}:1521/sidname")
            == f"oracle://scott:***@{url_host}:1521/sidname"
        )
    except ModuleNotFoundError:
        pass

    assert (
        PasswordMasker.mask_db_url(
            f"oracle://scott:tiger@{url_host}:1521/sidname", use_urlparse=True
        )
        == f"oracle://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{url_host}:1521/sidname"
    )

    try:
        assert (
            PasswordMasker.mask_db_url("oracle+cx_oracle://scott:tiger@tnsname")
            == f"oracle+cx_oracle://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@tnsname"
        )
    except ModuleNotFoundError:
        pass
    assert (
        PasswordMasker.mask_db_url(
            "oracle+cx_oracle://scott:tiger@tnsname", use_urlparse=True
        )
        == f"oracle+cx_oracle://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@tnsname"
    )

    # Microsoft SQL Server
    # pyodbc (if installed in test environment)
    try:
        assert (
            PasswordMasker.mask_db_url("mssql+pyodbc://scott:tiger@mydsn")
            == "mssql+pyodbc://scott:***@mydsn"
        )
    except ModuleNotFoundError:
        pass
    assert (
        PasswordMasker.mask_db_url(
            "mssql+pyodbc://scott:tiger@mydsn", use_urlparse=True
        )
        == f"mssql+pyodbc://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@mydsn"
    )

    # pymssql (if installed in test environment)
    try:
        assert (
            PasswordMasker.mask_db_url(
                f"mssql+pymssql://scott:tiger@{db_hostname}:12345/dbname"
            )
            == f"mssql+pymssql://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:12345/dbname"
        )
    except ModuleNotFoundError:
        pass
    assert (
        PasswordMasker.mask_db_url(
            f"mssql+pymssql://scott:tiger@{db_hostname}:12345/dbname", use_urlparse=True
        )
        == f"mssql+pymssql://scott:{PasswordMasker.MASKED_PASSWORD_STRING}@{db_hostname}:12345/dbname"
    )

    # SQLite
    # relative path
    temp_dir = tmp_path / "sqllite_tests"
    temp_dir.mkdir()
    monkeypatch.chdir(temp_dir)
    assert (
        PasswordMasker.mask_db_url("sqlite:///something/foo.db")
        == "sqlite:///something/foo.db"
    )
    assert (
        PasswordMasker.mask_db_url("sqlite:///something/foo.db", use_urlparse=True)
        == "sqlite:///something/foo.db"
    )

    # absolute path
    # Unix/Mac - 4 initial slashes in total
    assert (
        PasswordMasker.mask_db_url("sqlite:////absolute/path/to/foo.db")
        == "sqlite:////absolute/path/to/foo.db"
    )
    assert (
        PasswordMasker.mask_db_url(
            "sqlite:////absolute/path/to/foo.db", use_urlparse=True
        )
        == "sqlite:////absolute/path/to/foo.db"
    )

    # Windows
    assert (
        PasswordMasker.mask_db_url("sqlite:///C:\\path\\to\\foo.db")
        == "sqlite:///C:\\path\\to\\foo.db"
    )
    assert (
        PasswordMasker.mask_db_url("sqlite:///C:\\path\\to\\foo.db", use_urlparse=True)
        == "sqlite:///C:\\path\\to\\foo.db"
    )

    # Windows alternative using raw string
    assert (
        PasswordMasker.mask_db_url(r"sqlite:///C:\path\to\foo.db")
        == r"sqlite:///C:\path\to\foo.db"
    )
    assert (
        PasswordMasker.mask_db_url(r"sqlite:///C:\path\to\foo.db", use_urlparse=True)
        == r"sqlite:///C:\path\to\foo.db"
    )

    # in-memory
    assert PasswordMasker.mask_db_url("sqlite://") == "sqlite://"
    assert PasswordMasker.mask_db_url("sqlite://", use_urlparse=True) == "sqlite://"


def test_sanitize_config_azure_blob_store():
    azure_url: str = "DefaultEndpointsProtocol=https;AccountName=iamname;AccountKey=i_am_account_key;EndpointSuffix=core.windows.net"
    assert (
        PasswordMasker.mask_db_url(azure_url)
        == "DefaultEndpointsProtocol=https;AccountName=iamname;AccountKey=***;EndpointSuffix=core.windows.net"
    )

    azure_wrong_url: str = "DefaultEndpointsProtocol=i_dont_work;AccountName=iamname;AccountKey=i_am_account_key;EndpointSuffix=core.windows.net"
    with pytest.raises(StoreConfigurationError):
        PasswordMasker.mask_db_url(azure_wrong_url)

    azure_missing_fields: str = "DefaultEndpointsProtocol=i_dont_work;AccountName=iamname;EndpointSuffix=core.windows.net"
    with pytest.raises(StoreConfigurationError):
        PasswordMasker.mask_db_url(azure_missing_fields)


def test_sanitize_config_raises_exception_with_bad_input(
    basic_data_context_config,
):

    # expect that an Exception is raised if something other than a dict is passed
    with pytest.raises(TypeError):
        PasswordMasker.sanitize_config(basic_data_context_config)


def test_sanitize_config_doesnt_change_config_without_datasources(
    basic_data_context_config_dict,
):

    # expect no change without datasources
    config_without_creds = PasswordMasker.sanitize_config(
        basic_data_context_config_dict
    )
    assert config_without_creds == basic_data_context_config_dict


@pytest.mark.cloud
def test_sanitize_config_masks_cloud_store_backend_access_tokens(
    data_context_config_dict_with_cloud_backed_stores, ge_cloud_access_token
):

    # test that cloud store backend tokens have been properly masked
    config_with_creds_in_stores = PasswordMasker.sanitize_config(
        data_context_config_dict_with_cloud_backed_stores
    )
    for name, store_config in config_with_creds_in_stores["stores"].items():

        if (
            not store_config.get("store_backend")
            or not store_config["store_backend"].get("ge_cloud_credentials")
            or not store_config["store_backend"]["ge_cloud_credentials"].get(
                "access_token"
            )
        ):
            # a field in store_config["store_backend"]["ge_cloud_credentials"]["access_token"]
            # doesn't exist, so we expect this config to be unchanged
            assert (
                store_config
                == data_context_config_dict_with_cloud_backed_stores["stores"][name]
            )
        else:
            # check that the original token exists
            assert (
                data_context_config_dict_with_cloud_backed_stores["stores"][name][
                    "store_backend"
                ]["ge_cloud_credentials"]["access_token"]
                == ge_cloud_access_token
            )
            # expect that the GX Cloud token has been obscured
            assert (
                store_config["store_backend"]["ge_cloud_credentials"]["access_token"]
                != ge_cloud_access_token
            )


def test_sanitize_config_masks_execution_engine_connection_strings(
    data_context_config_dict_with_datasources, conn_string_password
):

    # test that datasource credentials have been properly masked
    unaltered_datasources = data_context_config_dict_with_datasources["datasources"]
    config_with_creds_masked = PasswordMasker.sanitize_config(
        data_context_config_dict_with_datasources
    )
    masked_datasources = config_with_creds_masked["datasources"]

    # iterate through the processed datasources and check for correctness
    for name, processed_config in masked_datasources.items():

        # check if processed_config["execution_engine"]["connection_string"] exists
        if processed_config.get("execution_engine") and processed_config[
            "execution_engine"
        ].get("connection_string"):

            # check if the connection string contains a password
            if (
                conn_string_password
                in unaltered_datasources[name]["execution_engine"]["connection_string"]
            ):
                # it does contain a password, so make sure its masked
                assert (
                    conn_string_password
                    not in processed_config["execution_engine"]["connection_string"]
                )
            else:
                # it doesn't contain a password, so make sure it's unaltered
                assert processed_config == unaltered_datasources[name]

        # processed_config either doesn't have an `execution_engine` field,
        # or a `connection_string` field
        else:
            # expect this config to be unaltered
            assert processed_config == unaltered_datasources[name]


def test_sanitize_config_with_arbitrarily_nested_sensitive_keys():

    # base case - this config should pass through unaffected
    config = {
        "some_field": "and a value",
        "some_other_field": {"password": "expect this to be found"},
    }
    config_copy = safe_deep_copy(config)
    res = PasswordMasker.sanitize_config(config_copy)
    assert res != config
    assert res["some_other_field"]["password"] == PasswordMasker.MASKED_PASSWORD_STRING


def test_sanitize_config_with_password_field():

    # this case has a password field inside a credentials dict - expect it to be masked
    config = {"credentials": {"password": "my-super-duper-secure-passphrase-123"}}
    config_copy = safe_deep_copy(config)
    res = PasswordMasker.sanitize_config(config_copy)
    assert res != config
    assert res["credentials"]["password"] == PasswordMasker.MASKED_PASSWORD_STRING


def test_sanitize_config_with_url_field(
    conn_string_with_embedded_password, conn_string_password
):

    # this case has a url field inside a credentials dict - expect the password inside
    # of it to be masked
    config = {"credentials": {"url": conn_string_with_embedded_password}}
    config_copy = safe_deep_copy(config)
    res = PasswordMasker.sanitize_config(config_copy)
    assert res != config
    assert conn_string_password not in res["credentials"]["url"]
    assert PasswordMasker.MASKED_PASSWORD_STRING in res["credentials"]["url"]


def test_sanitize_config_with_nested_url_field(
    conn_string_password, conn_string_with_embedded_password
):

    # this case has a connection string in an execution_engine dict
    config = {
        "execution_engine": {"connection_string": conn_string_with_embedded_password}
    }
    config_copy = safe_deep_copy(config)
    res = PasswordMasker.sanitize_config(config_copy)
    assert res != config
    assert conn_string_password not in res["execution_engine"]["connection_string"]
    assert (
        PasswordMasker.MASKED_PASSWORD_STRING
        in res["execution_engine"]["connection_string"]
    )


def test_sanitize_config_regardless_of_parent_key():

    # expect this config still be masked
    config = {
        "some_field": "and a value",
        "some_other_field": {"access_token": "but this won't be found"},
    }
    config_copy = safe_deep_copy(config)
    res = PasswordMasker.sanitize_config(config_copy)
    assert res != config
    assert (
        res["some_other_field"]["access_token"] == PasswordMasker.MASKED_PASSWORD_STRING
    )


@pytest.mark.cloud
def test_sanitize_config_masks_cloud_access_token(ge_cloud_access_token):

    # expect the access token to be found and masked
    config = {
        "store_backend": {
            "ge_cloud_credentials": {"access_token": ge_cloud_access_token}
        }
    }
    config_copy = safe_deep_copy(config)
    res = PasswordMasker.sanitize_config(config_copy)
    assert res != config
    assert (
        res["store_backend"]["ge_cloud_credentials"]["access_token"]
        == PasswordMasker.MASKED_PASSWORD_STRING
    )


def test_sanitize_config_works_with_list():
    config = {"some_key": [{"access_token": "12345"}]}
    config_copy = safe_deep_copy(config)
    res = PasswordMasker.sanitize_config(config_copy)
    assert res != config
    assert res["some_key"][0]["access_token"] == PasswordMasker.MASKED_PASSWORD_STRING


def test_parse_substitution_variable():
    """
    What does this test and why?
    Ensure parse_substitution_variable works as expected.
    Returns:

    """
    assert parse_substitution_variable("${SOME_VAR}") == "SOME_VAR"
    assert parse_substitution_variable("$SOME_VAR") == "SOME_VAR"
    assert parse_substitution_variable("SOME_STRING") is None
    assert parse_substitution_variable("SOME_$TRING") is None
    assert parse_substitution_variable("${some_var}") == "some_var"
    assert parse_substitution_variable("$some_var") == "some_var"
    assert parse_substitution_variable("some_string") is None
    assert parse_substitution_variable("some_$tring") is None
    assert parse_substitution_variable("${SOME_$TRING}") is None
    assert parse_substitution_variable("$SOME_$TRING") == "SOME_"



from typing import TYPE_CHECKING, Any, Dict, List, Optional

from great_expectations.rule_based_profiler.config import ParameterBuilderConfig
from great_expectations.rule_based_profiler.data_assistant import DataAssistant
from great_expectations.rule_based_profiler.data_assistant_result import (
    DataAssistantResult,
    VolumeDataAssistantResult,
)
from great_expectations.rule_based_profiler.domain_builder import (
    CategoricalColumnDomainBuilder,
    TableDomainBuilder,
)
from great_expectations.rule_based_profiler.expectation_configuration_builder import (
    DefaultExpectationConfigurationBuilder,
    ExpectationConfigurationBuilder,
)
from great_expectations.rule_based_profiler.helpers.cardinality_checker import (
    CardinalityLimitMode,
)
from great_expectations.rule_based_profiler.parameter_builder import (
    ParameterBuilder,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.parameter_container import (
    DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
    FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY,
    FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER,
    FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY,
    VARIABLES_KEY,
)
from great_expectations.rule_based_profiler.rule import Rule
from great_expectations.validator.validator import Validator  # noqa: TCH001

if TYPE_CHECKING:
    from great_expectations.rule_based_profiler.domain_builder import DomainBuilder


class VolumeDataAssistant(DataAssistant):
    """
    VolumeDataAssistant provides exploration and validation of "Data Volume" aspects of specified data Batch objects.
    """

    __alias__: str = "volume"

    def __init__(
        self,
        name: str,
        validator: Validator,
    ) -> None:
        super().__init__(
            name=name,
            validator=validator,
        )

    def get_variables(self) -> Optional[Dict[str, Any]]:
        """
        Returns:
            Optional "variables" configuration attribute name/value pairs (overrides), commonly-used in Builder objects.
        """
        return None

    def get_rules(self) -> Optional[List[Rule]]:
        """
        Returns:
            Optional custom list of "Rule" objects implementing particular "DataAssistant" functionality.
        """
        table_rule: Rule = self._build_table_rule()
        categorical_columns_rule: Rule = self._build_categorical_columns_rule()

        return [
            table_rule,
            categorical_columns_rule,
        ]

    def _build_data_assistant_result(
        self, data_assistant_result: DataAssistantResult
    ) -> DataAssistantResult:
        return VolumeDataAssistantResult(
            _batch_id_to_batch_identifier_display_name_map=data_assistant_result._batch_id_to_batch_identifier_display_name_map,
            profiler_config=data_assistant_result.profiler_config,
            profiler_execution_time=data_assistant_result.profiler_execution_time,
            rule_domain_builder_execution_time=data_assistant_result.rule_domain_builder_execution_time,
            rule_execution_time=data_assistant_result.rule_execution_time,
            metrics_by_domain=data_assistant_result.metrics_by_domain,
            expectation_configurations=data_assistant_result.expectation_configurations,
            citation=data_assistant_result.citation,
            _usage_statistics_handler=data_assistant_result._usage_statistics_handler,
        )

    @staticmethod
    def _build_table_rule() -> Rule:
        """
        This method builds "Rule" object focused on emitting "ExpectationConfiguration" objects for table "Domain" type.
        """
        # Step-1: Instantiate "TableDomainBuilder" object.

        table_domain_builder: DomainBuilder = TableDomainBuilder(
            data_context=None,
        )

        # Step-2: Declare "ParameterBuilder" for every metric of interest.

        table_row_count_metric_multi_batch_parameter_builder_for_metrics: ParameterBuilder = (
            DataAssistant.commonly_used_parameter_builders.get_table_row_count_metric_multi_batch_parameter_builder()
        )

        # Step-3: Declare "ParameterBuilder" for every "validation" need in "ExpectationConfigurationBuilder" objects.

        evaluation_parameter_builder_configs: Optional[List[ParameterBuilderConfig]] = [
            ParameterBuilderConfig(
                **table_row_count_metric_multi_batch_parameter_builder_for_metrics.to_json_dict()
            ),
        ]
        table_row_count_range_parameter_builder_for_validations: ParameterBuilder = DataAssistant.commonly_used_parameter_builders.build_numeric_metric_range_multi_batch_parameter_builder(
            metric_name=None,
            suffix=None,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
        )

        # Step-4: Pass "validation" "ParameterBuilderConfig" objects to every "DefaultExpectationConfigurationBuilder", responsible for emitting "ExpectationConfiguration" (with specified "expectation_type").

        validation_parameter_builder_configs: Optional[List[ParameterBuilderConfig]] = [
            ParameterBuilderConfig(
                **table_row_count_range_parameter_builder_for_validations.to_json_dict(),
            ),
        ]
        expect_table_row_count_to_be_between_expectation_configuration_builder = DefaultExpectationConfigurationBuilder(
            expectation_type="expect_table_row_count_to_be_between",
            validation_parameter_builder_configs=validation_parameter_builder_configs,
            min_value=f"{table_row_count_range_parameter_builder_for_validations.json_serialized_fully_qualified_parameter_name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY}[0]",
            max_value=f"{table_row_count_range_parameter_builder_for_validations.json_serialized_fully_qualified_parameter_name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY}[1]",
            meta={
                "profiler_details": f"{table_row_count_range_parameter_builder_for_validations.json_serialized_fully_qualified_parameter_name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY}",
            },
        )

        # Step-5: Instantiate and return "Rule" object, comprised of "variables", "domain_builder", "parameter_builders", and "expectation_configuration_builders" components.

        variables: dict = {
            "false_positive_rate": 0.05,
            "estimator": "bootstrap",
            "n_resamples": 9999,
            "random_seed": None,
            "quantile_statistic_interpolation_method": "nearest",
            "quantile_bias_correction": False,
            "quantile_bias_std_error_ratio_threshold": None,
            "include_estimator_samples_histogram_in_details": False,
            "truncate_values": {
                "lower_bound": 0,
                "upper_bound": None,
            },
            "round_decimals": 0,
        }
        parameter_builders: List[ParameterBuilder] = [
            table_row_count_metric_multi_batch_parameter_builder_for_metrics,
        ]
        expectation_configuration_builders: List[ExpectationConfigurationBuilder] = [
            expect_table_row_count_to_be_between_expectation_configuration_builder,
        ]
        rule = Rule(
            name="table_rule",
            variables=variables,
            domain_builder=table_domain_builder,
            parameter_builders=parameter_builders,
            expectation_configuration_builders=expectation_configuration_builders,
        )

        return rule

    @staticmethod
    def _build_categorical_columns_rule() -> Rule:
        """
        This method builds "Rule" object focused on emitting "ExpectationConfiguration" objects for categorical columns.
        """
        # Step-1: Instantiate "CategoricalColumnDomainBuilder" for selecting columns containing "FEW" discrete values.

        categorical_column_type_domain_builder: DomainBuilder = (
            CategoricalColumnDomainBuilder(
                include_column_names=None,
                exclude_column_names=None,
                include_column_name_suffixes=None,
                exclude_column_name_suffixes=None,
                semantic_type_filter_module_name=None,
                semantic_type_filter_class_name=None,
                include_semantic_types=None,
                exclude_semantic_types=None,
                allowed_semantic_types_passthrough=None,
                cardinality_limit_mode=f"{VARIABLES_KEY}cardinality_limit_mode",
                max_unique_values=None,
                max_proportion_unique=None,
                data_context=None,
            )
        )

        # Step-2: Declare "ParameterBuilder" for every metric of interest.

        column_distinct_values_count_metric_multi_batch_parameter_builder_for_metrics: ParameterBuilder = (
            DataAssistant.commonly_used_parameter_builders.get_column_distinct_values_count_metric_multi_batch_parameter_builder()
        )

        # Step-3: Declare "ParameterBuilder" for every "validation" need in "ExpectationConfigurationBuilder" objects.

        evaluation_parameter_builder_configs: Optional[List[ParameterBuilderConfig]] = [
            ParameterBuilderConfig(
                **column_distinct_values_count_metric_multi_batch_parameter_builder_for_metrics.to_json_dict()
            ),
        ]
        column_distinct_values_count_range_parameter_builder_for_validations: ParameterBuilder = DataAssistant.commonly_used_parameter_builders.build_numeric_metric_range_multi_batch_parameter_builder(
            metric_name=None,
            suffix=None,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
        )

        # Step-4: Pass "validation" "ParameterBuilderConfig" objects to every "DefaultExpectationConfigurationBuilder", responsible for emitting "ExpectationConfiguration" (with specified "expectation_type").

        validation_parameter_builder_configs: Optional[List[ParameterBuilderConfig]] = [
            ParameterBuilderConfig(
                **column_distinct_values_count_range_parameter_builder_for_validations.to_json_dict(),
            ),
        ]
        expect_column_unique_value_count_to_be_between_expectation_configuration_builder = DefaultExpectationConfigurationBuilder(
            expectation_type="expect_column_unique_value_count_to_be_between",
            validation_parameter_builder_configs=validation_parameter_builder_configs,
            column=f"{DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}column",
            min_value=f"{column_distinct_values_count_range_parameter_builder_for_validations.json_serialized_fully_qualified_parameter_name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY}[0]",
            max_value=f"{column_distinct_values_count_range_parameter_builder_for_validations.json_serialized_fully_qualified_parameter_name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY}[1]",
            strict_min=f"{VARIABLES_KEY}strict_min",
            strict_max=f"{VARIABLES_KEY}strict_max",
            meta={
                "profiler_details": f"{column_distinct_values_count_range_parameter_builder_for_validations.json_serialized_fully_qualified_parameter_name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY}",
            },
        )

        # Step-5: Instantiate and return "Rule" object, comprised of "variables", "domain_builder", "parameter_builders", and "expectation_configuration_builders" components.

        variables: dict = {
            "cardinality_limit_mode": CardinalityLimitMode.FEW.name,
            "mostly": 1.0,
            "strict_min": False,
            "strict_max": False,
            "false_positive_rate": 0.05,
            "estimator": "bootstrap",
            "n_resamples": 9999,
            "random_seed": None,
            "quantile_statistic_interpolation_method": "nearest",
            "quantile_bias_correction": False,
            "quantile_bias_std_error_ratio_threshold": None,
            "include_estimator_samples_histogram_in_details": False,
            "truncate_values": {
                "lower_bound": 0.0,
                "upper_bound": None,
            },
            "round_decimals": None,
        }
        parameter_builders: List[ParameterBuilder] = [
            column_distinct_values_count_metric_multi_batch_parameter_builder_for_metrics,
        ]
        expectation_configuration_builders: List[ExpectationConfigurationBuilder] = [
            expect_column_unique_value_count_to_be_between_expectation_configuration_builder,
        ]
        rule = Rule(
            name="categorical_columns_rule",
            variables=variables,
            domain_builder=categorical_column_type_domain_builder,
            parameter_builders=parameter_builders,
            expectation_configuration_builders=expectation_configuration_builders,
        )

        return rule



import os
from typing import List

from ruamel import yaml

import great_expectations as gx
from great_expectations.core.batch import Batch, BatchRequest

CREDENTIAL = os.getenv("AZURE_CREDENTIAL", "")

context = gx.get_context()

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/cloud/azure/pandas/configured_yaml_example.py datasource config">
datasource_yaml = f"""
name: my_azure_datasource
class_name: Datasource
execution_engine:
    class_name: PandasExecutionEngine
    azure_options:
        account_url: <YOUR_ACCOUNT_URL> # or `conn_str`
        credential: <YOUR_CREDENTIAL>   # if using a protected container
data_connectors:
    configured_data_connector_name:
        class_name: ConfiguredAssetAzureDataConnector
        azure_options:
            account_url: <YOUR_ACCOUNT_URL> # or `conn_str`
            credential: <YOUR_CREDENTIAL>   # if using a protected container
        container: <YOUR_AZURE_CONTAINER_HERE>
        name_starts_with: <CONTAINER_PATH_TO_DATA>
        default_regex:
            pattern: data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_(\\d{{4}})-(\\d{{2}})\\.csv
            group_names:
                - year
                - month
        assets:
            taxi_data:
"""
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
datasource_yaml = datasource_yaml.replace(
    "<YOUR_AZURE_CONTAINER_HERE>", "superconductive-public"
)
datasource_yaml = datasource_yaml.replace(
    "<CONTAINER_PATH_TO_DATA>", "data/taxi_yellow_tripdata_samples/"
)
datasource_yaml = datasource_yaml.replace(
    "<YOUR_ACCOUNT_URL>", "superconductivetesting.blob.core.windows.net"
)
datasource_yaml = datasource_yaml.replace("<YOUR_CREDENTIAL>", CREDENTIAL)

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/cloud/azure/pandas/configured_yaml_example.py test datasource">
context.test_yaml_config(datasource_yaml)
# </snippet>

context.add_datasource(**yaml.load(datasource_yaml))

# Here is a BatchRequest naming a data_asset
batch_request = BatchRequest(
    datasource_name="my_azure_datasource",
    data_connector_name="configured_data_connector_name",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name directly in the BatchRequest above.
batch_request.data_asset_name = "taxi_data"

context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())


# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)
assert [ds["name"] for ds in context.list_datasources()] == ["my_azure_datasource"]
assert set(
    context.get_available_data_asset_names()["my_azure_datasource"][
        "configured_data_connector_name"
    ]
) == {"taxi_data"}

batch_list: List[Batch] = context.get_batch_list(batch_request=batch_request)
assert len(batch_list) == 3

batch: Batch = batch_list[0]
assert batch.data.dataframe.shape[0] == 10000



from click.testing import CliRunner

from great_expectations.cli import cli


def test_validation_operator_with_new_api_raises_error():
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(cli, "--v3-api validation-operator")
    assert result.exit_code == 2
    assert "Error: No such command" in result.stderr
    assert ("'validation-operator'" in result.stderr) or (
        '"validation-operator"' in result.stderr
    )



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import jinja2
import nbformat

from great_expectations.core import ExpectationConfiguration  # noqa: TCH001
from great_expectations.core.batch import (
    BatchRequest,
    standardize_batch_request_display_ordering,
)
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.data_context.types.base import (
    NotebookConfig,  # noqa: TCH001
    NotebookTemplateConfig,  # noqa: TCH001
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.exceptions import (
    SuiteEditNotebookCustomTemplateModuleNotFoundError,
)
from great_expectations.render.renderer.notebook_renderer import BaseNotebookRenderer
from great_expectations.util import (
    deep_filter_properties_iterable,
    filter_properties_dict,
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )


class SuiteEditNotebookRenderer(BaseNotebookRenderer):
    """
    Render a notebook that can re-create or edit a suite.

    Use cases:
    - Make an easy path to edit a suite that a Profiler created.
    - Make it easy to edit a suite where only JSON exists.
    """

    def __init__(
        self,
        custom_templates_module: Optional[str] = None,
        header_markdown: Optional[NotebookTemplateConfig] = None,
        footer_markdown: Optional[NotebookTemplateConfig] = None,
        table_expectations_header_markdown: Optional[NotebookTemplateConfig] = None,
        column_expectations_header_markdown: Optional[NotebookTemplateConfig] = None,
        table_expectations_not_found_markdown: Optional[NotebookTemplateConfig] = None,
        column_expectations_not_found_markdown: Optional[NotebookTemplateConfig] = None,
        authoring_intro_markdown: Optional[NotebookTemplateConfig] = None,
        column_expectations_markdown: Optional[NotebookTemplateConfig] = None,
        header_code: Optional[NotebookTemplateConfig] = None,
        footer_code: Optional[NotebookTemplateConfig] = None,
        table_expectation_code: Optional[NotebookTemplateConfig] = None,
        column_expectation_code: Optional[NotebookTemplateConfig] = None,
        context: Optional[AbstractDataContext] = None,
    ) -> None:
        super().__init__(context=context)
        custom_loader: list = []

        if custom_templates_module:
            try:
                path_info_list: List[str] = custom_templates_module.rsplit(".", 1)
                package_name: str = path_info_list[0]
                package_path: str = path_info_list[1]
                custom_loader: list = [
                    jinja2.PackageLoader(
                        package_name=package_name,
                        package_path=package_path,
                    ),
                ]
            except ModuleNotFoundError as e:
                raise SuiteEditNotebookCustomTemplateModuleNotFoundError(
                    custom_templates_module
                ) from e

        loaders: list = custom_loader + [
            jinja2.PackageLoader(
                package_name="great_expectations.render.v3.notebook_assets",
                package_path="suite_edit",
            ),
        ]
        self.template_env = None

        self.template_env = jinja2.Environment(
            loader=jinja2.ChoiceLoader(loaders=loaders)
        )

        self.header_markdown = header_markdown
        self.footer_markdown = footer_markdown
        self.table_expectations_header_markdown = table_expectations_header_markdown
        self.column_expectations_header_markdown = column_expectations_header_markdown
        self.table_expectations_not_found_markdown = (
            table_expectations_not_found_markdown
        )
        self.column_expectations_not_found_markdown = (
            column_expectations_not_found_markdown
        )
        self.authoring_intro_markdown = authoring_intro_markdown
        self.column_expectations_markdown = column_expectations_markdown

        self.header_code = header_code
        self.footer_code = footer_code
        self.column_expectation_code = column_expectation_code
        self.table_expectation_code = table_expectation_code

    @staticmethod
    def from_data_context(data_context: AbstractDataContext):
        suite_edit_notebook_config: Optional[NotebookConfig] = None
        if data_context.notebooks and data_context.notebooks.get("suite_edit"):
            suite_edit_notebook_config = data_context.notebooks.get("suite_edit")

        if suite_edit_notebook_config:
            return instantiate_class_from_config(
                config=suite_edit_notebook_config.__dict__,
                runtime_environment={"context": data_context},
                config_defaults={},
            )
        return instantiate_class_from_config(
            config={
                "module_name": "great_expectations.render.renderer.v3.suite_edit_notebook_renderer",
                "class_name": "SuiteEditNotebookRenderer",
            },
            runtime_environment={"context": data_context},
            config_defaults={},
        )

    @classmethod
    def _get_expectations_by_column(
        cls, expectations: List[ExpectationConfiguration]
    ) -> Dict[str, List[ExpectationConfiguration]]:
        # TODO probably replace this with Suite logic at some point
        expectations_by_column: Dict[str, List[ExpectationConfiguration]] = {
            "table_expectations": []
        }

        expectation: ExpectationConfiguration
        column_name: str
        for expectation in expectations:
            if "column" in expectation["kwargs"]:
                column_name = expectation["kwargs"]["column"]
                if column_name not in expectations_by_column.keys():
                    expectations_by_column[column_name] = []
                expectations_by_column[column_name].append(expectation)
            else:
                expectations_by_column["table_expectations"].append(expectation)

        return expectations_by_column

    def render_with_overwrite(
        self,
        notebook_config: Optional[NotebookTemplateConfig],
        default_file_name: str,
        **default_kwargs,
    ):
        template: jinja2.Template
        rendered: str
        if notebook_config:
            template = self.template_env.get_template(name=notebook_config.file_name)
            rendered = template.render(
                **{**default_kwargs, **notebook_config.template_kwargs}
            )
        else:
            template = self.template_env.get_template(name=default_file_name)
            rendered = template.render(**default_kwargs)

        return rendered

    def add_header(
        self,
        suite_name: str,
        batch_request: Optional[Union[str, Dict[str, Any]]] = None,
    ) -> None:
        markdown: str = self.render_with_overwrite(
            notebook_config=self.header_markdown,
            default_file_name="HEADER.md",
            suite_name=suite_name,
        )
        self.add_markdown_cell(markdown=markdown)

        if batch_request is None:
            batch_request = {}

        code: str = self.render_with_overwrite(
            notebook_config=self.header_code,
            default_file_name="header.py.j2",
            suite_name=suite_name,
            batch_request=batch_request,
            env=os.environ,
        )
        self.add_code_cell(code=code, lint=True)

    def add_footer(
        self,
        batch_request: Optional[Union[str, Dict[str, Any]]] = None,
    ) -> None:
        markdown: str = self.render_with_overwrite(
            notebook_config=self.footer_markdown,
            default_file_name="FOOTER.md",
        )
        self.add_markdown_cell(markdown=markdown)
        code: str = self.render_with_overwrite(
            notebook_config=self.footer_code,
            default_file_name="footer.py.j2",
            batch_request=batch_request,
            env=os.environ,
        )
        self.add_code_cell(code=code)

    def add_expectation_cells_from_suite(
        self,
        expectations: List[ExpectationConfiguration],
        batch_request: Optional[Union[str, Dict[str, Any]]] = None,
    ) -> None:
        expectations_by_column: Dict[
            str, List[ExpectationConfiguration]
        ] = self._get_expectations_by_column(expectations=expectations)

        markdown: str

        markdown = self.render_with_overwrite(
            notebook_config=self.table_expectations_header_markdown,
            default_file_name="TABLE_EXPECTATIONS_HEADER.md",
        )
        self.add_markdown_cell(markdown=markdown)
        self._add_table_level_expectations(
            expectations_by_column=expectations_by_column, batch_request=batch_request
        )

        # Remove the table expectations since they are dealt with
        expectations_by_column.pop("table_expectations")

        markdown = self.render_with_overwrite(
            notebook_config=self.column_expectations_header_markdown,
            default_file_name="COLUMN_EXPECTATIONS_HEADER.md",
        )
        self.add_markdown_cell(markdown=markdown)
        self._add_column_level_expectations(
            expectations_by_column=expectations_by_column, batch_request=batch_request
        )

    def _add_table_level_expectations(
        self,
        expectations_by_column: Dict[str, List[ExpectationConfiguration]],
        batch_request: Optional[Union[str, Dict[str, Any]]] = None,
    ) -> None:
        if not expectations_by_column["table_expectations"]:
            markdown: str = self.render_with_overwrite(
                notebook_config=self.table_expectations_not_found_markdown,
                default_file_name="TABLE_EXPECTATIONS_NOT_FOUND.md",
                batch_request=batch_request,
                env=os.environ,
            )
            self.add_markdown_cell(markdown=markdown)
            return

        expectation: ExpectationConfiguration
        for expectation in expectations_by_column["table_expectations"]:
            filter_properties_dict(
                properties=expectation["kwargs"], clean_falsy=True, inplace=True
            )
            code: str = self.render_with_overwrite(
                notebook_config=self.table_expectation_code,
                default_file_name="table_expectation.py.j2",
                expectation=expectation,
                batch_request=batch_request,
                env=os.environ,
                kwargs_string=self._build_kwargs_string(expectation=expectation),
                meta_args=self._build_meta_arguments(meta=expectation.meta),
            )
            self.add_code_cell(code=code, lint=True)

    def _add_column_level_expectations(
        self,
        expectations_by_column: Dict[str, List[ExpectationConfiguration]],
        batch_request: Optional[Union[str, Dict[str, Any]]] = None,
    ) -> None:
        if not expectations_by_column:
            markdown: str = self.render_with_overwrite(
                notebook_config=self.column_expectations_not_found_markdown,
                default_file_name="COLUMN_EXPECTATIONS_NOT_FOUND.md",
                batch_request=batch_request,
                env=os.environ,
            )
            self.add_markdown_cell(markdown=markdown)
            return

        column_name: str
        expectations: List[ExpectationConfiguration]
        for column_name, expectations in expectations_by_column.items():
            markdown: str = self.render_with_overwrite(
                notebook_config=self.column_expectations_markdown,
                default_file_name="COLUMN_EXPECTATIONS.md",
                column=column_name,
            )
            self.add_markdown_cell(markdown=markdown)

            expectation: ExpectationConfiguration
            for expectation in expectations:
                filter_properties_dict(
                    properties=expectation["kwargs"], clean_falsy=True, inplace=True
                )
                code: str = self.render_with_overwrite(
                    notebook_config=self.column_expectation_code,
                    default_file_name="column_expectation.py.j2",
                    expectation=expectation,
                    batch_request=batch_request,
                    env=os.environ,
                    kwargs_string=self._build_kwargs_string(expectation=expectation),
                    meta_args=self._build_meta_arguments(meta=expectation.meta),
                )
                self.add_code_cell(code=code, lint=True)

    @classmethod
    def _build_kwargs_string(cls, expectation: ExpectationConfiguration) -> str:
        kwargs: List[str] = []
        expectation_kwargs: dict = filter_properties_dict(
            properties=expectation["kwargs"], clean_falsy=True
        )
        for k, v in expectation_kwargs.items():
            if k == "column":
                # make the column a positional argument
                kwargs.insert(0, f"{k}='{v}'")
            elif isinstance(v, str):
                # Put strings in quotes
                kwargs.append(f"{k}='{v}'")
            else:
                # Pass other types as is
                kwargs.append(f"{k}={v}")

        return ", ".join(kwargs)

    @staticmethod
    def _build_meta_arguments(meta: Dict[str, str]) -> str:
        if not meta:
            return ""

        profiler: str = "BasicSuiteBuilderProfiler"
        if profiler in meta.keys():
            meta.pop(profiler)

        if meta.keys():
            return f", meta={meta}"

        return ""

    def render(self, **kwargs: dict) -> nbformat.NotebookNode:
        """
        Render a notebook dict from an expectation suite.
        """
        # noinspection PyTypeChecker
        suite: ExpectationSuite = kwargs.get("suite")
        batch_request: Optional[Union[str, Dict[str, Any]]] = kwargs.get(
            "batch_request"
        )

        if not isinstance(suite, ExpectationSuite):
            raise RuntimeWarning("render must be given an ExpectationSuite.")

        self._notebook = nbformat.v4.new_notebook()

        suite_name: str = suite.expectation_suite_name

        if (
            batch_request
            and isinstance(batch_request, dict)
            and BatchRequest(**batch_request)
        ):
            batch_request = standardize_batch_request_display_ordering(
                batch_request=batch_request
            )
        else:
            batch_request = None

        self.add_header(suite_name=suite_name, batch_request=batch_request)
        self.add_authoring_intro(batch_request=batch_request)
        self.add_expectation_cells_from_suite(
            expectations=suite.expectations, batch_request=batch_request
        )
        self.add_footer(batch_request=batch_request)

        return self._notebook

    def render_to_disk(self, notebook_file_path: str, **kwargs: dict) -> None:
        """
        Render a notebook to disk from an expectation suite.

        If batch_request dictionary is passed, its properties will override any found in suite citations.
        """
        # noinspection PyTypeChecker
        suite: ExpectationSuite = kwargs.get("suite")
        batch_request: Optional[Union[str, Dict[str, Any]]] = kwargs.get(
            "batch_request"
        )

        deep_filter_properties_iterable(
            properties=batch_request,
            inplace=True,
        )
        # noinspection PyTypeChecker
        self.render(
            suite=suite,
            batch_request=batch_request,
        )
        self.write_notebook_to_disk(
            notebook=self._notebook, notebook_file_path=notebook_file_path
        )

    def add_authoring_intro(
        self,
        batch_request: Optional[Union[str, Dict[str, Any]]] = None,
    ) -> None:
        markdown: str = self.render_with_overwrite(
            notebook_config=self.authoring_intro_markdown,
            default_file_name="AUTHORING_INTRO.md",
            batch_request=batch_request,
            env=os.environ,
        )
        self.add_markdown_cell(markdown=markdown)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List, Optional

from great_expectations.core import (
    ExpectationConfiguration,  # noqa: TCH001
    ExpectationSuite,  # noqa: TCH001
)
from great_expectations.core._docs_decorators import public_api
from great_expectations.core.domain import Domain  # noqa: TCH001
from great_expectations.core.usage_statistics.events import UsageStatsEvents
from great_expectations.core.usage_statistics.usage_statistics import (
    UsageStatisticsHandler,
    get_expectation_suite_usage_statistics,
    usage_statistics_enabled_method,
)
from great_expectations.core.util import convert_to_json_serializable
from great_expectations.rule_based_profiler.helpers.util import (
    get_or_create_expectation_suite,
)
from great_expectations.rule_based_profiler.parameter_container import (
    ParameterNode,  # noqa: TCH001
)
from great_expectations.types import SerializableDictDot

if TYPE_CHECKING:
    from great_expectations.alias_types import JSONValues


@public_api
@dataclass(frozen=True)
class RuleBasedProfilerResult(SerializableDictDot):
    """
    ``RuleBasedProfilerResult`` is an immutable ``dataclass`` object which holds the results of executing the ``RuleBasedProfiler.run()`` method.

    Properties represents the configuration of the Rule-Based Profiler (effective configuration if run via a DataAssistant or auto-initializing expectation), with all run-time overrides properly reconciled.

    Args:
        fully_qualified_parameter_names_by_domain:
            `dict` of `Domain` keys and a list of their parameter names.
        parameter_values_for_fully_qualified_parameter_names_by_domain:
            `dict` of `Domain` and nested `ParameterNode` mappings.
        expectation_configurations:
            List of `ExpectationConfiguration` objects.
        citation:
            `dict` of citations.

    """

    fully_qualified_parameter_names_by_domain: Dict[Domain, List[str]]
    parameter_values_for_fully_qualified_parameter_names_by_domain: Optional[
        Dict[Domain, Dict[str, ParameterNode]]
    ]
    expectation_configurations: List[ExpectationConfiguration]
    citation: dict
    rule_domain_builder_execution_time: Dict[str, float]
    rule_execution_time: Dict[str, float]
    # Reference to  "UsageStatisticsHandler" object for this "RuleBasedProfilerResult" object (if configured).
    _usage_statistics_handler: Optional[UsageStatisticsHandler] = field(default=None)

    def to_dict(self) -> dict:
        """
        Returns:
            This `RuleBasedProfilerResult` as dictionary (JSON-serializable for `RuleBasedProfilerResult` objects).
        """
        domain: Domain
        fully_qualified_parameter_names: List[str]
        parameter_values_for_fully_qualified_parameter_names: Dict[str, ParameterNode]
        expectation_configuration: ExpectationConfiguration
        parameter_values_for_fully_qualified_parameter_names_by_domain: Dict[
            Domain, Dict[str, ParameterNode]
        ] = (self.parameter_values_for_fully_qualified_parameter_names_by_domain or {})
        return {
            "fully_qualified_parameter_names_by_domain": [
                {
                    "domain_id": domain.id,
                    "domain": domain.to_json_dict(),
                    "fully_qualified_parameter_names": convert_to_json_serializable(
                        data=fully_qualified_parameter_names
                    ),
                }
                for domain, fully_qualified_parameter_names in self.fully_qualified_parameter_names_by_domain.items()
            ],
            "parameter_values_for_fully_qualified_parameter_names_by_domain": [
                {
                    "domain_id": domain.id,
                    "domain": domain.to_json_dict(),
                    "parameter_values_for_fully_qualified_parameter_names": convert_to_json_serializable(
                        data=parameter_values_for_fully_qualified_parameter_names
                    ),
                }
                for domain, parameter_values_for_fully_qualified_parameter_names in parameter_values_for_fully_qualified_parameter_names_by_domain.items()
            ],
            "expectation_configurations": [
                expectation_configuration.to_json_dict()
                for expectation_configuration in self.expectation_configurations
            ],
            "citation": self.citation,
        }

    @public_api
    def to_json_dict(self) -> dict[str, JSONValues]:
        """
        Returns the `RuleBasedProfilerResult` as a JSON-serializable dictionary.

        Returns:
            Dictionary containing only JSON compatible python primitives.
        """
        return self.to_dict()

    @public_api
    @usage_statistics_enabled_method(
        event_name=UsageStatsEvents.RULE_BASED_PROFILER_RESULT_GET_EXPECTATION_SUITE,
        args_payload_fn=get_expectation_suite_usage_statistics,
    )
    def get_expectation_suite(self, expectation_suite_name: str) -> ExpectationSuite:
        """
        Retrieve the `ExpectationSuite` generated during the `RuleBasedProfiler` run.

        Args:
            expectation_suite_name: The name of the desired `ExpectationSuite`.

        Returns:
            `ExpectationSuite`
        """
        expectation_suite: ExpectationSuite = get_or_create_expectation_suite(
            data_context=None,
            expectation_suite=None,
            expectation_suite_name=expectation_suite_name,
            component_name=self.__class__.__name__,
            persist=False,
        )
        expectation_suite.add_expectation_configurations(
            expectation_configurations=self.expectation_configurations,
            send_usage_event=False,
            match_type="domain",
            overwrite_existing=True,
        )
        expectation_suite.add_citation(
            **self.citation,
        )
        return expectation_suite
